# -*- coding:utf-8 -*-
'''
@Project  : fypy

@File     : jsonpro.py

@Modify Time : 2022/8/11 15:34

@Author : Lee

@Version : 1.0

@Description :
对json\ascii\二进制文件进行读写操作

'''
import sys
import os
import numpy as np
import json



def readjson(jsonname, buffering=None, encoding=None, **kw):
    '''
    读取JSON文件

    Parameters
    ----------
    jsonname: str
        JSON文件名
    kwargs

    Returns
    -------
        dict
        json文件内容以dict形式返回
    '''

    fp = open(jsonname, 'r')
    dict_info_in = json.load(fp, **kw)

    fp.close()

    return dict_info_in

def writejson(jsonname, dict_info, indent=4, chinese=False):

    if chinese :
        ensure_ascii = False
    else:
        ensure_ascii = True

    fp = open(jsonname, 'w')
    json.dump(dict_info, fp, indent=indent, ensure_ascii=ensure_ascii)
    fp.close()


def readbinary(filename, shape, dtype=np.float32, offset=0, encoding='utf-8'):
    '''
    读取二进制文件

    Parameters
    ----------
    filename: str
        输入文件名
    shape: list or tuple
        获取的维度大小
    dtype: numpy.dtype
        数据类型
    offset: int
        偏移量
    encoding: str
        编码类型
    Returns
    -------
        numpy.array
    '''

    count = 1
    if os.path.isfile(filename) :
        for i in shape :
            count *= i
        data = np.fromfile(filename, dtype=dtype, count= count,offset=offset)

    return data.reshape(shape)


def writebinary(filename, data, overwrite=1, offset=0, encoding='utf-8'):
    '''
    写入二进制文件

    Parameters
    ----------
    filename: str
        输出文件名
    shape: list or tuple
        获取的维度大小
    overwrite: int
        是否覆盖、新建文件, 非0为新建
    offset: int
        偏移量
    encoding: str
        编码类型
    Returns
    -------
        numpy.array
    '''

    data = np.array(data)

    if overwrite:
        fp = open(filename, 'wb')
    else:
        fp = open(filename, 'ab')

    fp.seek(offset)
    data.tofile(fp)

    fp.close()

def readascii(filename, dtype=float, comments='#', delimiter=None,
                     converters=None, skiprows=0, usecols=None, unpack=False,
                     ndmin=0, encoding='bytes', max_rows=None):
    '''
    Load data from a text file.

    Each row in the text file must have the same number of values.

    Parameters
    ----------
    fname : file, str, or pathlib.Path
        File, filename, or generator to read.  If the filename extension is
        ``.gz`` or ``.bz2``, the file is first decompressed. Note that
        generators should return byte strings.
    dtype : data-type, optional
        Data-type of the resulting array; default: float.  If this is a
        structured data-type, the resulting array will be 1-dimensional, and
        each row will be interpreted as an element of the array.  In this
        case, the number of columns used must match the number of fields in
        the data-type.
    comments : str or sequence of str, optional
        The characters or list of characters used to indicate the start of a
        comment. None implies no comments. For backwards compatibility, byte
        strings will be decoded as 'latin1'. The default is '#'.
    delimiter : str, optional
        The string used to separate values. For backwards compatibility, byte
        strings will be decoded as 'latin1'. The default is whitespace.
    converters : dict, optional
        A dictionary mapping column number to a function that will parse the
        column string into the desired value.  E.g., if column 0 is a date
        string: ``converters = {0: datestr2num}``.  Converters can also be
        used to provide a default value for missing data (but see also
        `genfromtxt`): ``converters = {3: lambda s: float(s.strip() or 0)}``.
        Default: None.
    skiprows : int, optional
        Skip the first `skiprows` lines, including comments; default: 0.
    usecols : int or sequence, optional
        Which columns to read, with 0 being the first. For example,
        ``usecols = (1,4,5)`` will extract the 2nd, 5th and 6th columns.
        The default, None, results in all columns being read.

        .. versionchanged:: 1.11.0
            When a single column has to be read it is possible to use
            an integer instead of a tuple. E.g ``usecols = 3`` reads the
            fourth column the same way as ``usecols = (3,)`` would.
    unpack : bool, optional
        If True, the returned array is transposed, so that arguments may be
        unpacked using ``x, y, z = loadtxt(...)``.  When used with a
        structured data-type, arrays are returned for each field.
        Default is False.
    ndmin : int, optional
        The returned array will have at least `ndmin` dimensions.
        Otherwise mono-dimensional axes will be squeezed.
        Legal values: 0 (default), 1 or 2.

        .. versionadded:: 1.6.0
    encoding : str, optional
        Encoding used to decode the inputfile. Does not apply to input streams.
        The special value 'bytes' enables backward compatibility workarounds
        that ensures you receive byte arrays as results if possible and passes
        'latin1' encoded strings to converters. Override this value to receive
        unicode arrays and pass strings as input to converters.  If set to None
        the system default is used. The default value is 'bytes'.

        .. versionadded:: 1.14.0
    max_rows : int, optional
        Read `max_rows` lines of content after `skiprows` lines. The default
        is to read all the lines.

        .. versionadded:: 1.16.0
    ${ARRAY_FUNCTION_LIKE}

        .. versionadded:: 1.20.0

    Returns
    -------
    out : ndarray
        Data read from the text file.

    See Also
    --------
    load, fromstring, fromregex
    genfromtxt : Load data with missing values handled as specified.
    scipy.io.loadmat : reads MATLAB data files

    Notes
    -----
    This function aims to be a fast reader for simply formatted files.  The
    `genfromtxt` function provides more sophisticated handling of, e.g.,
    lines with missing values.
    '''
    return  np.loadtxt(filename, dtype=dtype, comments=comments, delimiter=delimiter,
                       converters=converters, skiprows=skiprows, usecols=usecols, unpack=unpack,
                       ndmin=ndmin, encoding=encoding, max_rows=max_rows)

def writeascii(filename, data,  fmt='%.18e', delimiter=' ', newline='\n', header='',
               footer='', comments='# ', encoding=None):
    '''
    Save an array to a text file.

    Parameters
    ----------
    fname : filename or file handle
        If the filename ends in ``.gz``, the file is automatically saved in
        compressed gzip format.  `loadtxt` understands gzipped files
        transparently.
    X : 1D or 2D array_like
        Data to be saved to a text file.
    fmt : str or sequence of strs, optional
        A single format (%10.5f), a sequence of formats, or a
        multi-format string, e.g. 'Iteration %d -- %10.5f', in which
        case `delimiter` is ignored. For complex `X`, the legal options
        for `fmt` are:

        * a single specifier, `fmt='%.4e'`, resulting in numbers formatted
          like `' (%s+%sj)' % (fmt, fmt)`
        * a full string specifying every real and imaginary part, e.g.
          `' %.4e %+.4ej %.4e %+.4ej %.4e %+.4ej'` for 3 columns
        * a list of specifiers, one per column - in this case, the real
          and imaginary part must have separate specifiers,
          e.g. `['%.3e + %.3ej', '(%.15e%+.15ej)']` for 2 columns
    delimiter : str, optional
        String or character separating columns.
    newline : str, optional
        String or character separating lines.

        .. versionadded:: 1.5.0
    header : str, optional
        String that will be written at the beginning of the file.

        .. versionadded:: 1.7.0
    footer : str, optional
        String that will be written at the end of the file.

        .. versionadded:: 1.7.0
    comments : str, optional
        String that will be prepended to the ``header`` and ``footer`` strings,
        to mark them as comments. Default: '# ',  as expected by e.g.
        ``numpy.loadtxt``.

        .. versionadded:: 1.7.0
    encoding : {None, str}, optional
        Encoding used to encode the outputfile. Does not apply to output
        streams. If the encoding is something other than 'bytes' or 'latin1'
        you will not be able to load the file in NumPy versions < 1.14. Default
        is 'latin1'.

        .. versionadded:: 1.14.0


    See Also
    --------
    save : Save an array to a binary file in NumPy ``.npy`` format
    savez : Save several arrays into an uncompressed ``.npz`` archive
    savez_compressed : Save several arrays into a compressed ``.npz`` archive

    Notes
    -----
    Further explanation of the `fmt` parameter
    (``%[flag]width[.precision]specifier``):

    flags:
        ``-`` : left justify

        ``+`` : Forces to precede result with + or -.

        ``0`` : Left pad the number with zeros instead of space (see width).

    width:
        Minimum number of characters to be printed. The value is not truncated
        if it has more characters.

    precision:
        - For integer specifiers (eg. ``d,i,o,x``), the minimum number of
          digits.
        - For ``e, E`` and ``f`` specifiers, the number of digits to print
          after the decimal point.
        - For ``g`` and ``G``, the maximum number of significant digits.
        - For ``s``, the maximum number of characters.

    specifiers:
        ``c`` : character

        ``d`` or ``i`` : signed decimal integer

        ``e`` or ``E`` : scientific notation with ``e`` or ``E``.

        ``f`` : decimal floating point

        ``g,G`` : use the shorter of ``e,E`` or ``f``

        ``o`` : signed octal

        ``s`` : string of characters

        ``u`` : unsigned decimal integer

        ``x,X`` : unsigned hexadecimal integer

    This explanation of ``fmt`` is not complete, for an exhaustive
    specification see [1]_.

    References
    ----------
    .. [1] `Format Specification Mini-Language
           <https://docs.python.org/library/string.html#format-specification-mini-language>`_,
           Python Documentation.
    '''
    data = np.array(data)

    np.savetxt(filename, data, fmt, delimiter, newline, header,
               footer, comments, encoding)

def loadarray(file, mmap_mode=None, allow_pickle=False, fix_imports=True,
              encoding='ASCII'):
    ''' 读取numpy保存输出的.npy文件 '''
    return np.load(file, mmap_mode, allow_pickle, fix_imports, encoding)


def savearray(filename, data, allow_pickle=True, fix_imports=True):
    ''' 输出.npy文件 '''
    return  np.save(filename, data, allow_pickle, fix_imports)





# -*- coding:utf-8 -*-
'''
@Project  : fypy
@File     : __init__.py
@Modify Time      @Author    @Version    
--------------    -------    --------    
2022/7/19 14:51      Lee       1.0         
@Description
------------------------------------
 
'''

from .hdf4pro import readhdf4, readhdf4sdsattrs, readhdf4fileattrs
from .h4toh5 import h4toh5
from .hdfpro import *
from .ncpro import *
from .tifpro import *
from .jsonpro import *

from .nc2tif import nc2tif



# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : AtmCorr_Landsat.py

@Modify Time :  2022/12/23 16:06   

@Author      : Lee    

@Version     : 1.0   

@Description :

'''
import os
import sys
import numpy as np
import datetime
import re
import glob
from osgeo import gdal
from Py6S import *
from .AtmCorr import AtmCorr

class AtmCorr_Landsat(AtmCorr):

    def __init__(self,  metafile):
        super(AtmCorr_Landsat, self).__init__()

        metadata = {}

        with open(metafile, 'r') as fp :
            lines = fp.readlines()

        for line in lines :
            if 'GROUP' in line :
                continue

            if '=' in line :
                line = line.replace('\n','')
                lineinfo = line.split('=')
                key = lineinfo[0].replace(' ', '')
                # value = lineinfo[1].replace(' ', '')
                value = lineinfo[1]
                metadata[key] = value

        self.metadata = metadata

    def FLAASH(self, nowdate, srcfile, BandId=None, outname=None, fillvalue=65535):

        if BandId is None :
            BandId = int((os.path.basename(srcfile).split('.')[0])[-1])

        #捕捉打开数据出错异常
        try:
            IDataSet = gdal.Open(srcfile)
        except Exception as e:
            print("文件%S打开失败" % srcfile)

        #获取行列号
        cols = IDataSet.RasterXSize
        rows = IDataSet.RasterYSize
        ImgBand = IDataSet.GetRasterBand(1)
        ImgRasterData = ImgBand.ReadAsArray(0, 0, cols, rows)

        trans = IDataSet.GetGeoTransform()
        prj = IDataSet.GetProjection()

        #辐射校正
        radiance = self.GetRadiance(BandId, ImgRasterData, self.metadata)
        #大气校正
        corrdata = self.GetAtmCorr(nowdate, BandId, radiance, self.metadata, dem=0)

        print('第%s波段大气校正完成' %BandId)

        return corrdata, trans, prj


    def GetAtmCorr(self, nowdate, BandId, radiance, metadata, dem):
        '''

        :param nowdate:
        :param BandId:
        :param radiance:
        :param metadata:
        :param dem:
        :return:
        '''
        # minwl, maxwl, response = atmc.getresp(respfile)

        # 中心经纬度
        point1lat = float(metadata['CORNER_UL_LAT_PRODUCT'])
        point1lon = float(metadata['CORNER_UL_LON_PRODUCT'])
        point2lat = float(metadata['CORNER_UR_LAT_PRODUCT'])
        point2lon = float(metadata['CORNER_UR_LON_PRODUCT'])
        point3lat = float(metadata['CORNER_LL_LAT_PRODUCT'])
        point3lon = float(metadata['CORNER_LL_LON_PRODUCT'])
        point4lat = float(metadata['CORNER_LR_LAT_PRODUCT'])
        point4lon = float(metadata['CORNER_LR_LON_PRODUCT'])

        sLongitude = (point1lon + point2lon + point3lon + point4lon) / 4
        sLatitude = (point1lat + point2lat + point3lat + point4lat) / 4

        sunz = 90-float(metadata['SUN_ELEVATION'])
        suna =    float(metadata['SUN_AZIMUTH'])

        # 通过研究去区的范围去求DEM高度。
        # pointUL = dict()
        # pointDR = dict()
        # pointUL["lat"] = point1lat
        # pointUL["lon"] = point1lon
        # pointDR["lat"] = point4lat
        # pointDR["lon"] = point2lon

        # 校正波段（根据波段名称）
        if BandId == 1:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B1

        elif BandId == 2:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B2

        elif BandId == 3:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B3

        elif BandId == 4:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B4

        elif BandId == 5:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B5

        elif BandId == 6:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B6

        elif BandId == 7:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B7

        elif BandId == 8:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B8

        elif BandId == 9:
            wavelength = PredefinedWavelengths.LANDSAT_OLI_B9

        self.set_geom(nowdate, sunz=sunz, suna=suna, satz=0, sata=0)
        self.set_atm(sLatitude=sLatitude, nowdate=nowdate)
        self.set_aer()
        self.set_vis()
        self.set_altitude(dem=dem)
        self.set_resp(wavelength=wavelength)
        a, b, c  = self.corrCoeff()
        corrdata = self.corrImage(radiance, a, b, c)

        return np.array(corrdata)


    # 逐波段辐射定标
    def GetRadiance(self, BandId, data, metadata, fillvalue=65535):
        ''' 计算辐射亮度参数：DN -> radiance '''

        Gain = float(metadata['RADIANCE_MULT_BAND_%d' %(BandId)])
        Bias = float(metadata['RADIANCE_ADD_BAND_%d' %(BandId)])

        RaCal = np.where(data>0 ,Gain * data + Bias, fillvalue)

        return RaCal

    def GetReflectance(self, BandId, data, metadata, fillvalue=65535):
        ''' 计算反射率：DN -> Reflectance '''

        Gain = float(metadata['REFLECTANCE_MULT_BAND_%d' %(BandId)])
        Bias = float(metadata['REFLECTANCE_ADD_BAND_%d' %(BandId)])

        RaCal = np.where(data>0 ,Gain * data + Bias, fillvalue)

        return RaCal



# -*- coding:utf-8 -*-
'''
@Project  : fypy
@File     : __init__.py
@Modify Time      @Author    @Version    
--------------    -------    --------    
2022/7/20 11:29      Lee       1.0         
@Description
------------------------------------
 
'''

from .version import get_versions
__version__ = get_versions()['version']
del get_versions

# __version__ = "1.1.12"
# version_tuple = (1, 1, 12)
# try:
#     from fypy.version import version as __version__  # noqa
# except ModuleNotFoundError:
#     raise ModuleNotFoundError(
#         "No module named fypy.version. This could mean "
#         "you didn't install 'fypy' properly. Try reinstalling ('pip "
#         "install').")


# try:
#     __import__('pkg_resources').declare_namespace(__name__)
# except ImportError:
#     pass  # must not have setuptools



# -*- coding:utf-8 -*-
'''
@Project  : fypy

@File     : hdf4pro.py

@Modify Time : 2022/8/11 15:34

@Author : Lee

@Version : 1.0

@Description :
对hdf4文件进行读写

'''
import os

def readhdf4(h4file, sdsname, dictsdsattrs=None, dictfileattrs=None):
    ''' 读取hdf4数据集 '''
    data = None

    from pyhdf import SD
    if not os.path.isfile(h4file):
        print('%s is not exist, will continue...' %(h4file))
        return data

    fp4 = SD.SD(h4file, SD.SDC.READ)

    if sdsname in fp4.datasets().keys() :
        try:
            sds4id = fp4.select(sdsname)
            data = sds4id[:]
        except BaseException as e:
            print(e)
    fp4.end()

    if dictfileattrs is not None and dictsdsattrs is not None :
        return data, dictfileattrs, dictsdsattrs
    elif dictfileattrs is not None and dictsdsattrs is None :
        return data, dictfileattrs
    elif dictfileattrs is None and dictsdsattrs is not None :
        return data, dictsdsattrs
    else:
        return data


def readhdf4sdsattrs(h4file, sdsname):
    ''' 读取hdf4文件数据集属性信息'''
    from pyhdf import SD

    sdsattrs = {}
    if not os.path.isfile(h4file):
        print('%s is not exist, will continue...' %(h4file))
        return sdsattrs
    try:
        fp4 = SD.SD(h4file, SD.SDC.READ)
        sds4id = fp4.select(sdsname)
        attrs = sds4id.attributes(full=1)
        for key in sds4id.attributes() :
            sdsattrs[key] =  attrs[key][0]
    except BaseException as e:
        print(e)

    return sdsattrs


def readhdf4fileattrs(h4file):
    ''' 读取hdf4文件属性'''
    from pyhdf import SD

    fileattrs = {}

    if not os.path.isfile(h4file):
        print('%s is not exist, will continue...' %(h4file))
        return fileattrs

    fp4 = SD.SD(h4file, SD.SDC.READ)

    attrs = fp4.attributes(full=1)

    for item in fp4.attributes().keys():
        try:
            fileattrs[item] = attrs[item][0]
        except BaseException as e:
            print(e)

    return fileattrs


# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : fy4pro.py

@Modify Time :  2022/11/10 14:45

@Author      : Lee

@Version     : 1.0

@Description :

'''
import os
import sys
import numpy as np
from osgeo import gdal, osr
from fypy.tools import readhdf_fileinfo

class fy4pro(object) :

    def __init__(self):
        pass


    def nom2gll(self, data, outname=None, bbox=None, subpoint=104.7, resolution=0.04,
                fillvalue=None, dstSRS='EPSG:4326'):
        '''
        将DISK 投影转换成等经纬投影

        Parameters
        ----------
        data : numpy.array
            输入数据
        outname: str, optional
            输出文件名
        bbox : tuple, optional
            output bounds as (minX, minY, maxX, maxY) in target SRS
        subpoint: float
            星下点
        resolution: float
            数据图像的分辨率，单位为degree
        fillvalue : float
            数据填充值
        dstSRS

        Returns
        -------

        '''
        im_data = np.array(data, dtype=np.float32)
        dtype = self._gettype(im_data.dtype)
        if len(im_data.shape) == 3:
            im_bands, im_height, im_width = im_data.shape
        elif len(im_data.shape) == 2:
            im_bands, (im_height, im_width) = 1,im_data.shape
        else:
            im_bands, (im_height, im_width) = 1,im_data.shape

        Driver = gdal.GetDriverByName('MEM')
        memDs = Driver.Create('', im_height, im_width, im_bands, dtype)

        # 写入数据
        if im_bands == 1:
            memDs.GetRasterBand(1).WriteArray(im_data)
        else:
            for i in range(im_bands):
                memDs.GetRasterBand(i+1).WriteArray(im_data[i])

        srs = osr.SpatialReference()
        srs.ImportFromProj4('+proj=geos +h=35785863 +a=6378137.0 +b=6356752.3 +lon_0={subpoint} +no_defs'.format(
            subpoint=subpoint ))
        memDs.SetProjection(srs.ExportToWkt())
        memDs.SetGeoTransform([-5496000, resolution*100*1000, 0, 5496000, 0, -resolution*100*1000])
        if outname is None :
            warpDs = gdal.Warp('', memDs, format='MEM', dstSRS=dstSRS,
                               outputBounds=bbox, xRes=resolution, yRes=resolution,
                               dstNodata=fillvalue)
        else:
            warpDs = gdal.Warp(outname, memDs, format='GTiff', dstSRS=dstSRS,
                               outputBounds=bbox, xRes=resolution, yRes=resolution,
                               srcNodata=fillvalue, dstNodata=fillvalue, creationOptions=["COMPRESS=LZW"])
        if warpDs is None :
            return None

        data_GLL = warpDs.ReadAsArray(0, 0, warpDs.RasterXSize, warpDs.RasterYSize)
        del warpDs

        return data_GLL

    def getL1Data(self, filename, bandID=1, fillvalue=65535):
        '''
        读取FY4 L1数据，并完成辐射定标
        Parameters
        ----------
        filename: str
            L1数据文件名
        bandID : int
            波段索引，从1开始

        Returns
        -------
            numpy.array
            辐射定标转换后的ref或bt
        '''
        if not os.path.isfile(filename) :
            print('文件不存在【%s】' %(filename))
            return None

        import h5py

        # 转换到区域的行列号（考虑去除图像偏移）
        fileinfo = readhdf_fileinfo(filename)
        Begin_Line_Number = fileinfo['Begin Line Number'][0]
        End_Line_Number = fileinfo['End Line Number'][0]
        Begin_Pixel_Number = fileinfo['Begin Pixel Number'][0]
        End_Pixel_Number = fileinfo['End Pixel Number'][0]

        # data = np.zeros(shape=(End_Line_Number+1, End_Pixel_Number+1),dtype=np.float32)

        fp = h5py.File(filename, 'r')
        cal = fp['CALChannel%02d' %(bandID)][:]
        dn = fp['NOMChannel%02d' %(bandID)][:]
        fp.close()

        flag = dn>=len(cal)
        dn[flag] = 0

        data = cal[dn]
        data[flag] = fillvalue
        # data[Begin_Line_Number:End_Line_Number+1, Begin_Pixel_Number:End_Pixel_Number+1] = cal[dn]

        return data


    def getGEOData(self, filename, sdsname):
        import h5py
        fp = h5py.File(filename, 'r')
        data1 = fp[sdsname][:]
        fp.close()

        # 转换到区域的行列号（考虑去除图像偏移）
        fileinfo = readhdf_fileinfo(filename)
        Begin_Line_Number = fileinfo['Begin Line Number'][0]
        End_Line_Number = fileinfo['End Line Number'][0]
        Begin_Pixel_Number = fileinfo['Begin Pixel Number'][0]
        End_Pixel_Number = fileinfo['End Pixel Number'][0]

        # data = np.zeros(shape=(End_Line_Number+1, End_Pixel_Number+1),dtype=np.float32)
        #
        # data[Begin_Line_Number:End_Line_Number+1, Begin_Pixel_Number:End_Pixel_Number+1] = data1

        return data1

    def _gettype(self, datatype):
        ''' 根据numpy的数据类型，匹配GDAL中的数据类型 '''

        if datatype == np.byte or datatype == np.uint8:
            return gdal.GDT_Byte
        elif datatype == np.uint16 :
            return gdal.GDT_UInt16
        elif datatype == np.int16 :
            return gdal.GDT_Int16
        elif datatype == np.uint32 :
            return gdal.GDT_UInt32
        elif datatype == np.int32 :
            return gdal.GDT_Int32
        elif datatype == np.float32 or datatype.str in ['>f4', '<f4']:
            return gdal.GDT_Float32
        elif datatype == np.float64 or datatype.str in ['>f8', '<f8']:
            return gdal.GDT_Float64
        else:
            return gdal.GDT_Unknown





# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : __init__.py

@Modify Time :  2022/11/10 14:45

@Author      : Lee

@Version     : 1.0

@Description :

'''
import os
import sys
import numpy as np
import datetime

from osgeo import gdal, ogr, osr

from fypy.tools import readhdf, readhdf_fileinfo, writehdf, writetiff

from .config import FY3Block10CoefX, FY3Block10CoefY, ProdInfo

class fy3pro(object) :

    def __init__(self):
        pass

    def Calibration(self, filename, sdsname):

        # EV_250_Aggr.1KM_RefSB     1~4         # 0.47, 0.55, 0.65, 0.865,
        # EV_1KM_RefSB              5~19        # 1.38, 1.64, 2.13, 0.412,
        #                                         0.443, 0.49, 0.555, 0.67,
        #                                         0.709, 0.746, 0.865, 0.905,
        #                                         0.936, 0.94, 1.03,
        # EV_1KM_Emissive           20~23       # 3.796, 4.046, 7.233, 8.56,
        # EV_250_Aggr.1KM_Emissive  24~25       # 10.714, 11.948

        if 'EV_1KM_Emissive' in sdsname :
            data = readhdf(filename, sdsname)
            fileinfo = readhdf_fileinfo(filename)
            wave_length = fileinfo['Effect_Center_WaveLength']
            return self.calemiss(data, 10000.0/wave_length[np.arange(20, 24)-1])

        elif 'EV_250_Aggr.1KM_Emissive' in sdsname :
            data = readhdf(filename, sdsname)
            fileinfo = readhdf_fileinfo(filename)
            wave_length = fileinfo['Effect_Center_WaveLength']
            return self.calemiss(data, 10000.0/wave_length[np.arange(24, 26)-1])

        elif 'EV_250_Aggr.1KM_RefSB' in sdsname :
            data = readhdf(filename, sdsname)
            coef = readhdf(filename, '/Calibration/VIS_Cal_Coeff')

            return self.calref(data, coef[np.arange(1, 5)-1, :])

        elif 'EV_1KM_RefSB' in sdsname :
            data = readhdf(filename, sdsname)
            coef = readhdf(filename, '/Calibration/VIS_Cal_Coeff')

            return self.calref(data, coef[np.arange(5, 20)-1, :])

        else:
            tmpdata, sdsinfo = readhdf(filename, sdsname,dictsdsinfo={})
            data = tmpdata.copy()
            if 'Slope' in sdsinfo and 'Intercept' in sdsinfo:
                data = data * sdsinfo['Slope'] + sdsinfo['Intercept']

            if 'valid_range' in sdsinfo :
                data[(tmpdata < sdsinfo['valid_range'][0]) | (tmpdata > sdsinfo['valid_range'][1])] = np.nan

            return data

    def calref(self, dn, coef):
        ''' 可见光通道定标 '''
        ref = np.full_like(dn, fill_value=np.nan, dtype=np.float32)
        for i in range(dn.shape[0]) :
            ref[i] = coef[i, 0] + coef[i, 1] * dn[i] + coef[i, 2] * dn[i] * dn[i]

            ref[i, dn[i] == 65535] = np.nan
            ref[i, ref[i]<=0] = 0

        return ref

    def calemiss(self, dn, wavenum):
        ''' 红外波段定标 '''
        temp = dn * 0.01
        temp[(temp<=0) | (temp >= 600.0)] = np.nan

        bt = np.full_like(temp, fill_value=np.nan, dtype=np.float32)

        for i in range(temp.shape[0]) :
            bt[i] = self.planck_r2t(temp[i, :, :], wavenum[i])

        return bt

    def planck_r2t(self, rad, wn):
        '''
        普朗克函数：将辐射值转成亮温（K）

        Parameters
        ----------
        rad : numpy.narray
            mW/(m2.cm-1.sr)
        wn : float or numpy.narray
            wave number(cm^-1)

        Returns
        -------
            numpy.narray
            bright temperature
            units: K
        '''
        # 普朗克系数
        Radiation_C1 = 1.191042953E-5
        Radiation_C2 = 1.4387774

        bt = (Radiation_C2 * wn / np.log(Radiation_C1 * wn * wn * wn / (rad)+1.0))

        return bt


class FY3Block10():

    def __init__(self, SatID, InstID, productName, filelist, sdsname, resolution, outname=None):

        self.resolution = resolution
        self.SatID = SatID
        self.InstID = InstID
        self.SDSName = sdsname

        if isinstance(filelist, list):
            self._MultipleFiles(filelist, productName, outname=outname)
        elif isinstance(filelist, str):
            self._SingleFiles(filelist, productName, outname=outname)

    def _SingleFiles(self, filename, productName, outname=None) :
        '''单个数据产品投影拼接'''
        basename = os.path.basename(filename)
        names = basename.split('_')
        blockID = names[2]

        lat = FY3Block10CoefY[blockID[0:2]]
        lon = FY3Block10CoefX[blockID[2:4]]

        mtrans = [lon, self.resolution, 0,
                  lat, 0, -self.resolution]

        Dict_DataSets = self.GetData(filename, productName)

        ds = writetiff(outname=outname, data=Dict_DataSets[self.SDSName],
                       im_geotrans=mtrans, fillvalue=self.fillvalue)

        return ds

    def _MultipleFiles(self, filelist, productName, outname=None) :
        '''多个数据产品投影拼接'''
        countfile = len(filelist)
        if countfile == 0 :
            return None
        elif countfile == 1 :
            self._SingleFiles(filelist[0], productName)
        else:
            all_ds = []

            for filename in filelist :
                dset = self._SingleFiles(filename, productName)
                all_ds.append(dset)

            if outname is None :
                format = 'MEM'
                outname=''
                creationOptions=[]
            else:
                format = 'GTiff'
                creationOptions=["COMPRESS=LZW"]

            ds = gdal.Warp(outname, all_ds, dstSRS='EPSG:4326', format=format,
                           xRes=self.resolution, yRes=self.resolution, dstNodata=self.fillvalue,
                           srcNodata=self.fillvalue, resampleAlg=gdal.GRA_NearestNeighbour,
                           creationOptions=creationOptions)

            return ds

    def getBoundingBox(self, filelist):

        tileindex = []

        for filename in filelist :
            basename = os.path.basename(filename)
            names = basename.split('_')
            blockID = names[2]

            lat = FY3Block10CoefY[blockID[0:2]]
            lon = FY3Block10CoefX[blockID[2:4]]

            tileindex.append([lat, lon])

        tileindex = np.array(tileindex)
        maxY = np.nanmax(tileindex[:,0])
        minY = np.nanmin(tileindex[:,0])

        maxX = np.nanmax(tileindex[:,1])
        minX = np.nanmin(tileindex[:,1])

        # mtrans = [minX, self.resolution, 0,
        #           minY, 0, self.resolution]

        extent = [minX, maxX+10, minY, maxY+10]

        return extent

    def GetData(self, filename, productName):

        Dict_DataSets = {}

        try:
            if productName in ProdInfo[self.SatID][self.InstID]:
                val, sdsinfo = readhdf(filename, self.SDSName, dictsdsinfo={})
                if 'FillValue' in sdsinfo :

                    if isinstance(sdsinfo['FillValue'], np.ndarray) :
                        self.fillvalue = sdsinfo['FillValue'][0]
                    else:
                        self.fillvalue = sdsinfo['FillValue']

                else:
                    self.fillvalue = 65535

                if 'valid_range' in sdsinfo :
                    vmin = sdsinfo['valid_range'][0]
                    vmax = sdsinfo['valid_range'][1]
                    fillflag = (val<vmin) | (val>vmax)

                if 'Slope' in sdsinfo and 'Intercept' in sdsinfo :
                    val = val * sdsinfo['Slope'] + sdsinfo['Intercept']

                val[fillflag] = self.fillvalue
                if 'valid_range' in sdsinfo :
                    vmin = sdsinfo['valid_range'][0]
                    vmax = sdsinfo['valid_range'][1]
                    fillflag = (val<vmin) | (val>vmax)
                    # val[fillflag] = self.fillvalue

                if productName == 'CLM':
                    data_u8 = np.uint8(val)
                    data_bit = np.unpackbits(data_u8).reshape(data_u8.size, 8)
                    bit21 = data_bit[:, 5] * 4 + data_bit[:, 6] * 2 + data_bit[:, 7]
                    val = bit21.reshape(val.shape)
                    # 001 = Cloudy （1）
                    # 011 = Uncertain （3）
                    # 101 = Probably  Clear （5）
                    # 111 = Confident  Clear （7）
                    val[(val != 1) & (val != 3) & (val != 5) & (val != 7)] = 255
                    val[val==1] = 0
                    val[val==3] = 1
                    val[val==5] = 2
                    val[val==7] = 3
                    self.fillvalue = 255

                key = self.SDSName.replace(' ', '_')
                if '/' in key :
                    key = key.split('/')[-1]
                Dict_DataSets[key] = val
            else:
                print('产品【%s】不在绘图要求列表之内' %(productName))
        except BaseException as  e :
            print('读取【%s】：%s失败！！！' %(productName, filename))

        return Dict_DataSets

class FY3Orbit() :
    '''针对FY-3卫星数据，包含MERSI 5分钟块、
    MWRI 升降轨、MWHS、MWTS等整圈轨道数据的投影、拼接、裁剪'''
    def __init__(self, srcdata, srclat, srclon, dstfile=None,
                 vmin=None, vmax=None, resolution=0.01,
                 minX=None, maxX=None, minY=None, maxY=None,  # 需要投影的区域范围
                 resampleAlg=gdal.GRIORA_NearestNeighbour,
                 srcNodata=-999.0, dstNodata=None, dstSRS="EPSG:4326"):
        '''

        Parameters
        ----------
        srcdata :
        srclat :
        srclon :
        dstfile :
        vmin :
        vmax :
        resolution :
        minX :
        maxX :
        minY :
        maxY :
        resampleAlg :
        srcNodata :
        dstNodata :
        dstSRS :
        '''
        self.TempFile = []

        if dstfile is None :
            srcfile = './tempfile_%s.temp' %(datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S'))
        else:
            srcfile = dstfile

        if dstNodata is None :
            dstNodata = srcNodata

        flag = (srclon > 180) | (srclon < -180) \
             | (srclat > 90)  | (srclat < -90)
        srclon[flag] = np.nan
        srclat[flag] = np.nan
        # srcdata[flag] = np.nan

        # 获取投影数据的范围
        if maxX is None :
            maxX = np.nanmax(srclon)

        if minX is None :
            minX = np.nanmin(srclon)

        if maxY is None :
            maxY = np.nanmax(srclat)

        if minY is None :
            minY = np.nanmin(srclat)

        if vmin is None :
            vmin = np.nanmin(srcdata)

        if vmax is None :
            vmax = np.nanmax(srcdata)

        data = np.array(srcdata).copy()


        if vmax is not None and vmin is not None :
            data[(srcdata < vmin) | (srcdata > vmax)] = srcNodata

        data[np.isnan(data)] = srcNodata

        temphdf = os.path.splitext(srcfile)[0] + '.hdf'
        self.TempFile.append(temphdf)
        # 创建临时的数据文件
        writehdf(temphdf, 'data', data, overwrite=1)
        writehdf(temphdf, 'lon', srclon, overwrite=0)
        writehdf(temphdf, 'lat', srclat, overwrite=0)

        layer = self._GetSourceInfo(temphdf, 'data')

        vrtFile = os.path.splitext(srcfile)[0] + '.vrt'
        self.TempFile.append(vrtFile)

        self._createVrt(vrtFile, temphdf, layer, '/lon', '/lat')

        if dstfile is None :
            format = 'MEM'
            dstfile = None
        else:
            format = 'GTiff'

        geoData = gdal.Warp(dstfile, vrtFile,
                            format=format,  geoloc=True,
                            dstSRS=dstSRS,  resampleAlg=resampleAlg,
                            srcNodata= srcNodata, dstNodata=dstNodata,
                            outputBounds=(minX, minY, maxX, maxY),  # (minX, minY, maxX, maxY)
                            xRes=resolution, yRes=resolution,
                            creationOptions=["COMPRESS=LZW"])

        if geoData == None:
            print('处理失败')
            return None

        width = geoData.RasterXSize #栅格矩阵的列数
        height = geoData.RasterYSize #栅格矩阵的行数
        im_bands = geoData.RasterCount #波段数

        im_data = geoData.ReadAsArray(0, 0, width, height)#获取数据
        im_geotrans = geoData.GetGeoTransform()#获取仿射矩阵信息
        im_proj = geoData.GetProjection()#获取投影信息

        if dstfile is not None :
            print('完成投影转换{}\n'.format(srcfile))

        if vmax is not None and vmin is not None :
            im_data[(im_data < vmin) | (im_data > vmax)] = dstNodata

    def _GetSourceInfo(self, filename, sdsname):
        src_ds = gdal.Open(filename)
        layers = src_ds.GetSubDatasets()

        # 获取sdsname所在的图层栅格索引
        src_raster = self.GetLayer(layers, sdsname)
        if src_raster is None :
            raise Exception('数据集【%s】不在文件中【%s】' %(sdsname, filename))

        return src_raster

    def GetLayer(self, layers, sdsname):
        '''
        获取指定的图层的索引名
        :param layers: tuple
        :return: str
        '''

        if sdsname:
            for layer in layers :
                l_name = layer[0].split(':')[-1].replace('"','')
                # print(self.sdsname, l_name)
                if sdsname in l_name:
                    return layer[0]

        return None

    def _createVrt(self, vrtDir, srcfile, layer, srclon=None, srclat=None):

        if layer is None :
            gdal.Translate(vrtDir,
                           srcfile,
                           format='vrt')
        else:
            gdal.Translate(vrtDir,
                           layer,
                           format='vrt')

        lines = []
        with open(vrtDir, 'r') as f:
            for line in f:
                lines.append(line)
        lines.insert(1,'<Metadata domain="GEOLOCATION">\n')
        lines.insert(2,' <MDI key="LINE_OFFSET">1</MDI>\n')
        lines.insert(3, ' <MDI key="LINE_STEP">1</MDI>\n')
        lines.insert(4, ' <MDI key="PIXEL_OFFSET">1</MDI>\n')
        lines.insert(5, ' <MDI key="PIXEL_STEP">1</MDI>\n')
        lines.insert(6, ' <MDI key="SRS">GEOGCS["WGS84",'
                        'DATUM["WGS_1984",'
                        'SPHEROID["WGS84",6378137,298.257223563,'
                        'AUTHORITY["EPSG","7030"]],'
                        'TOWGS84[0,0,0,0,0,0,0],'
                        'AUTHORITY["EPSG","6326"]],'
                        'PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],'
                        'UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9108"]],'
                        'AUTHORITY["EPSG","4326"]]</MDI>\n')
        lines.insert(7, ' <MDI key="X_BAND">1</MDI>')
        lines.insert(8, ' <MDI key="X_DATASET">HDF5:"{geofile}":/{srclon}</MDI>\n'.format(
            geofile=srcfile, srclon=srclon))
        lines.insert(9, ' <MDI key="Y_BAND">1</MDI>\n')
        lines.insert(10, ' <MDI key="Y_DATASET">HDF5:"{geofile}":/{srclat}</MDI>\n'.format(
            geofile=srcfile, srclat=srclat))
        lines.insert(11, '</Metadata>\n')
        with open(vrtDir, 'w') as f:
            for line in lines:
                f.writelines(line)

    def __del__(self):
        '''
        清理临时文件
        :return:
        '''
        for item in self.TempFile :
            if os.path.isfile(item) :
                try:
                    os.remove(item)
                    # print(item)
                except BaseException as e:
                    print('删除%s失败' %(item))





# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : ahi8_read_hsd.py

@Modify Time :  2022/11/10 14:45

@Author      : Lee

@Version     : 1.0

@Description :

'''
import os
import time
import numpy as np
from datetime import datetime, timedelta
import tempfile
import bz2
import dask.array as da
import xarray as xr
from contextlib import closing
import shutil
from subprocess import Popen, PIPE
from io import BytesIO
from .geometry import AreaDefinition

try:
    from shutil import which
except ImportError:
    # python 2 - won't be used, but needed for mocking in tests
    which = None

AHI_CHANNEL_NAMES = ("1", "2", "3", "4", "5",
                     "6", "7", "8", "9", "10",
                     "11", "12", "13", "14", "15", "16")

# logger = logging.getLogger('ahi_hsd')

# Basic information block:
_BASIC_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                             ("blocklength", "<u2"),
                             ("total_number_of_hblocks", "<u2"),
                             ("byte_order", "u1"),
                             ("satellite", "S16"),
                             ("proc_center_name", "S16"),
                             ("observation_area", "S4"),
                             ("other_observation_info", "S2"),
                             ("observation_timeline", "<u2"),
                             ("observation_start_time", "f8"),
                             ("observation_end_time", "f8"),
                             ("file_creation_time", "f8"),
                             ("total_header_length", "<u4"),
                             ("total_data_length", "<u4"),
                             ("quality_flag1", "u1"),
                             ("quality_flag2", "u1"),
                             ("quality_flag3", "u1"),
                             ("quality_flag4", "u1"),
                             ("file_format_version", "S32"),
                             ("file_name", "S128"),
                             ("spare", "S40"),
                             ])

# Data information block
_DATA_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                            ("blocklength", "<u2"),
                            ("number_of_bits_per_pixel", "<u2"),
                            ("number_of_columns", "<u2"),
                            ("number_of_lines", "<u2"),
                            ("compression_flag_for_data", "u1"),
                            ("spare", "S40"),
                            ])

# Projection information block
# See footnote 2; LRIT/HRIT Global Specification Section 4.4, CGMS, 1999)
_PROJ_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                            ("blocklength", "<u2"),
                            ("sub_lon", "f8"),
                            ("CFAC", "<u4"),
                            ("LFAC", "<u4"),
                            ("COFF", "f4"),
                            ("LOFF", "f4"),
                            ("distance_from_earth_center", "f8"),
                            ("earth_equatorial_radius", "f8"),
                            ("earth_polar_radius", "f8"),
                            ("req2_rpol2_req2", "f8"),
                            ("rpol2_req2", "f8"),
                            ("req2_rpol2", "f8"),
                            ("coeff_for_sd", "f8"),
                            # Note: processing center use only:
                            ("resampling_types", "<i2"),
                            # Note: processing center use only:
                            ("resampling_size", "<i2"),
                            ("spare", "S40"),
                            ])

# Navigation information block
_NAV_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                           ("blocklength", "<u2"),
                           ("navigation_info_time", "f8"),
                           ("SSP_longitude", "f8"),
                           ("SSP_latitude", "f8"),
                           ("distance_earth_center_to_satellite", "f8"),
                           ("nadir_longitude", "f8"),
                           ("nadir_latitude", "f8"),
                           ("sun_position", "f8", (3,)),
                           ("moon_position", "f8", (3,)),
                           ("spare", "S40"),
                           ])

# Calibration information block
_CAL_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                           ("blocklength", "<u2"),
                           ("band_number", "<u2"),
                           ("central_wave_length", "f8"),
                           ("valid_number_of_bits_per_pixel", "<u2"),
                           ("count_value_error_pixels", "<u2"),
                           ("count_value_outside_scan_pixels", "<u2"),
                           ("gain_count2rad_conversion", "f8"),
                           ("offset_count2rad_conversion", "f8"),
                           ])

# Infrared band (Band No. 7 – 16)
# (Band No. 2 – 5: backup operation (See Table 4 bb))
_IRCAL_INFO_TYPE = np.dtype([("c0_rad2tb_conversion", "f8"),
                             ("c1_rad2tb_conversion", "f8"),
                             ("c2_rad2tb_conversion", "f8"),
                             ("c0_tb2rad_conversion", "f8"),
                             ("c1_tb2rad_conversion", "f8"),
                             ("c2_tb2rad_conversion", "f8"),
                             ("speed_of_light", "f8"),
                             ("planck_constant", "f8"),
                             ("boltzmann_constant", "f8"),
                             ("spare", "S40"),
                             ])

# Visible, near-infrared band (Band No. 1 – 6)
# (Band No. 1: backup operation (See Table 4 bb))
_VISCAL_INFO_TYPE = np.dtype([("coeff_rad2albedo_conversion", "f8"),
                              ("coeff_update_time", "f8"),
                              ("cali_gain_count2rad_conversion", "f8"),
                              ("cali_offset_count2rad_conversion", "f8"),
                              ("spare", "S80"),
                              ])

# 6 Inter-calibration information block
_INTER_CALIBRATION_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("gsics_calibration_intercept", "f8"),
    ("gsics_calibration_slope", "f8"),
    ("gsics_calibration_coeff_quadratic_term", "f8"),
    ("gsics_std_scn_radiance_bias", "f8"),
    ("gsics_std_scn_radiance_bias_uncertainty", "f8"),
    ("gsics_std_scn_radiance", "f8"),
    ("gsics_correction_starttime", "f8"),
    ("gsics_correction_endtime", "f8"),
    ("gsics_radiance_validity_upper_lim", "f4"),
    ("gsics_radiance_validity_lower_lim", "f4"),
    ("gsics_filename", "S128"),
    ("spare", "S56"),
])

# 7 Segment information block
_SEGMENT_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("total_number_of_segments", "u1"),
    ("segment_sequence_number", "u1"),
    ("first_line_number_of_image_segment", "u2"),
    ("spare", "S40"),
])

# 8 Navigation correction information block
_NAVIGATION_CORRECTION_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("center_column_of_rotation", "f4"),
    ("center_line_of_rotation", "f4"),
    ("amount_of_rotational_correction", "f8"),
    ("numof_correction_info_data", "<u2"),
])

# 9 Observation time information block
_OBS_TIME_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("number_of_observation_times", "<u2"),
])

# 10 Error information block
_ERROR_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u4"),
    ("number_of_error_info_data", "<u2"),
])

# 11 Spare block
_SPARE_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("spare", "S256")
])



class ahi8_read_hsd(object):

    def GetData(self, hsdname, segnum, segtotal):
        self.segment_number = segnum
        self.segtotal = segtotal
        self.data = None
        self.is_zipped = False

        self._unzipped = self.unzip_file(hsdname)
        if self._unzipped:
            # But if it is, set the filename to point to unzipped temp file
            self.is_zipped = True
            self.filename = self._unzipped
        else:
            self.filename = hsdname

        try:
            with open(self.filename) as fd:
                self.basic_info = np.fromfile(fd,
                                              dtype=_BASIC_INFO_TYPE,
                                              count=1)
                self.data_info = np.fromfile(fd,
                                             dtype=_DATA_INFO_TYPE,
                                             count=1)
                self.proj_info = np.fromfile(fd,
                                             dtype=_PROJ_INFO_TYPE,
                                             count=1)[0]
                self.nav_info = np.fromfile(fd,
                                            dtype=_NAV_INFO_TYPE,
                                            count=1)[0]
                fd.close()
        except BaseException as e :
            print(e)
            # os.remove(self.filename)
            return False

        self.platform_name = np2str(self.basic_info['satellite'])
        self.sensor = 'ahi'
        # self.segment_number = filename_info['segment']
        # self.total_segments = filename_info['total_segments']
        self.observation_area = np2str(self.basic_info['observation_area'])

        user_calibration=None
        calib_mode = 'NOMINAL'
        calib_mode_choices = ('NOMINAL', 'UPDATE')
        if calib_mode.upper() not in calib_mode_choices:
            raise ValueError('Invalid calibration mode: {}. Choose one of {}'.format(
                calib_mode, calib_mode_choices))

        self.calib_mode = calib_mode.upper()
        self.user_calibration = user_calibration

        with open(self.filename, "rb") as fp_:
            self._header = self._read_header(fp_)
            res = self._read_data(fp_, self._header)
            fp_.close()
        self.data = self.read_band(res)

        return True

    def read_band(self, res):

        res = self._mask_invalid(data=res, header=self._header)
        # Calibrate
        bandnum = self._header["block5"]['band_number'][0]
        if bandnum < 7:
            res = self.calibrate(res, 'reflectance', )
        else:
            res = self.calibrate(res, 'brightness_temperature')

        # Update metadata
        new_info = dict(
            # units=info['units'],
            # standard_name=info['standard_name'],
            # wavelength=info['wavelength'],
            resolution='resolution',
            # id=key,
            # name=key['name'],
            # scheduled_time=self.scheduled_time,
            platform_name=self.platform_name,
            sensor=self.sensor,
            satellite_longitude=float(self.nav_info['SSP_longitude']),
            satellite_latitude=float(self.nav_info['SSP_latitude']),
            satellite_altitude=float(self.nav_info['distance_earth_center_to_satellite'] -
                                     self.proj_info['earth_equatorial_radius']) * 1000,
            orbital_parameters={
                'projection_longitude': float(self.proj_info['sub_lon']),
                'projection_latitude': 0.,
                'projection_altitude': float(self.proj_info['distance_from_earth_center'] -
                                             self.proj_info['earth_equatorial_radius']) * 1000,
                # 'satellite_actual_longitude': actual_lon,
                # 'satellite_actual_latitude': actual_lat,
                # 'satellite_actual_altitude': actual_alt,
                'nadir_longitude': float(self.nav_info['nadir_longitude']),
                'nadir_latitude': float(self.nav_info['nadir_latitude'])}
        )
        res = xr.DataArray(res, attrs=new_info, dims=['y', 'x'])

        self.get_area_def()

        self.mask_space = True
        # Mask space pixels
        if self.mask_space:
            res = self._mask_space(res)

        return res

    def calibrate(self, data, calibration):
        """Calibrate the data."""
        tic = datetime.now()

        if calibration == 'counts':
            return data

        if calibration in ['radiance', 'reflectance', 'brightness_temperature']:
            data = self.convert_to_radiance(data)

        if calibration == 'reflectance':
            data = self._vis_calibrate(data)
        elif calibration == 'brightness_temperature':
            data = self._ir_calibrate(data)

        # print("Calibration time " + str(datetime.now() - tic))
        return data

    def convert_to_radiance(self, data):
        """Calibrate to radiance."""
        bnum = self._header["block5"]['band_number'][0]
        # Check calibration mode and select corresponding coefficients
        if self.calib_mode == "UPDATE" and bnum < 7:
            dn_gain = self._header['calibration']["cali_gain_count2rad_conversion"][0]
            dn_offset = self._header['calibration']["cali_offset_count2rad_conversion"][0]
            if dn_gain == 0 and dn_offset == 0:
                print("No valid updated coefficients, fall back to default values.")
                dn_gain = self._header["block5"]["gain_count2rad_conversion"][0]
                dn_offset = self._header["block5"]["offset_count2rad_conversion"][0]
        else:
            dn_gain = self._header["block5"]["gain_count2rad_conversion"][0]
            dn_offset = self._header["block5"]["offset_count2rad_conversion"][0]

        # Assume no user correction
        correction_type = None
        if isinstance(self.user_calibration, dict):
            # Check if we have DN correction coeffs
            if 'type' in self.user_calibration:
                correction_type = self.user_calibration['type']
            else:
                # If not, assume radiance correction
                correction_type = 'RAD'
            if correction_type == 'DN':
                # Replace file calibration with user calibration
                dn_gain, dn_offset = get_user_calibration_factors(self.band_name,
                                                                  self.user_calibration)
            elif correction_type == 'RAD':
                user_slope, user_offset = get_user_calibration_factors(self.band_name,
                                                                       self.user_calibration)

        data = (data * dn_gain + dn_offset).clip(0)
        # If using radiance correction factors from GSICS or similar, apply here
        if correction_type == 'RAD':
            data = apply_rad_correction(data, user_slope, user_offset)
        return data

    def _vis_calibrate(self, data):
        """Visible channel calibration only."""
        coeff = self._header["calibration"]["coeff_rad2albedo_conversion"]
        return (data * coeff * 100).clip(0)

    def _ir_calibrate(self, data):
        """IR calibration."""
        # No radiance -> no temperature
        data = da.where(data == 0, np.float32(np.nan), data)

        cwl = self._header['block5']["central_wave_length"][0] * 1e-6
        c__ = self._header['calibration']["speed_of_light"][0]
        h__ = self._header['calibration']["planck_constant"][0]
        k__ = self._header['calibration']["boltzmann_constant"][0]
        a__ = (h__ * c__) / (k__ * cwl)

        b__ = ((2 * h__ * c__ ** 2) / (data * 1.0e6 * cwl ** 5)) + 1

        Te_ = a__ / da.log(b__)

        c0_ = self._header['calibration']["c0_rad2tb_conversion"][0]
        c1_ = self._header['calibration']["c1_rad2tb_conversion"][0]
        c2_ = self._header['calibration']["c2_rad2tb_conversion"][0]

        return (c0_ + c1_ * Te_ + c2_ * Te_ ** 2).clip(0)

    def _read_header(self, fp_):
        """Read header."""
        header = {}

        fpos = 0
        header['block1'] = np.fromfile(
            fp_, dtype=_BASIC_INFO_TYPE, count=1)
        fpos = fpos + int(header['block1']['blocklength'])
        self._check_fpos(fp_, fpos, 0, 'block1')
        fp_.seek(fpos, 0)
        header["block2"] = np.fromfile(fp_, dtype=_DATA_INFO_TYPE, count=1)
        fpos = fpos + int(header['block2']['blocklength'])
        self._check_fpos(fp_, fpos, 0, 'block2')
        fp_.seek(fpos, 0)
        header["block3"] = np.fromfile(fp_, dtype=_PROJ_INFO_TYPE, count=1)
        fpos = fpos + int(header['block3']['blocklength'])
        self._check_fpos(fp_, fpos, 0, 'block3')
        fp_.seek(fpos, 0)
        header["block4"] = np.fromfile(fp_, dtype=_NAV_INFO_TYPE, count=1)
        fpos = fpos + int(header['block4']['blocklength'])
        self._check_fpos(fp_, fpos, 0, 'block4')
        fp_.seek(fpos, 0)
        header["block5"] = np.fromfile(fp_, dtype=_CAL_INFO_TYPE, count=1)
        # print("Band number = " + str(header["block5"]['band_number'][0]))
        # print('Time_interval: %s - %s', str(self.start_time), str(self.end_time))
        band_number = header["block5"]['band_number'][0]
        if band_number < 7:
            cal = np.fromfile(fp_, dtype=_VISCAL_INFO_TYPE, count=1)
        else:
            cal = np.fromfile(fp_, dtype=_IRCAL_INFO_TYPE, count=1)
        fpos = fpos + int(header['block5']['blocklength'])
        self._check_fpos(fp_, fpos, 0, 'block5')
        fp_.seek(fpos, 0)

        header['calibration'] = cal

        header["block6"] = np.fromfile(
            fp_, dtype=_INTER_CALIBRATION_INFO_TYPE, count=1)
        fpos = fpos + int(header['block6']['blocklength'])
        self._check_fpos(fp_, fpos, 0, 'block6')
        fp_.seek(fpos, 0)
        header["block7"] = np.fromfile(
            fp_, dtype=_SEGMENT_INFO_TYPE, count=1)
        fpos = fpos + int(header['block7']['blocklength'])
        self._check_fpos(fp_, fpos, 0, 'block7')
        fp_.seek(fpos, 0)
        header["block8"] = np.fromfile(
            fp_, dtype=_NAVIGATION_CORRECTION_INFO_TYPE, count=1)
        # 8 The navigation corrections:
        ncorrs = header["block8"]['numof_correction_info_data'][0]
        dtype = np.dtype([
            ("line_number_after_rotation", "<u2"),
            ("shift_amount_for_column_direction", "f4"),
            ("shift_amount_for_line_direction", "f4"),
        ])
        corrections = []
        for _i in range(ncorrs):
            corrections.append(np.fromfile(fp_, dtype=dtype, count=1))
        fpos = fpos + int(header['block8']['blocklength'])
        self._check_fpos(fp_, fpos, 40, 'block8')
        fp_.seek(fpos, 0)
        header['navigation_corrections'] = corrections
        header["block9"] = np.fromfile(fp_,
                                       dtype=_OBS_TIME_INFO_TYPE,
                                       count=1)
        numobstimes = header["block9"]['number_of_observation_times'][0]

        dtype = np.dtype([
            ("line_number", "<u2"),
            ("observation_time", "f8"),
        ])
        lines_and_times = []
        for _i in range(numobstimes):
            lines_and_times.append(np.fromfile(fp_,
                                               dtype=dtype,
                                               count=1))
        header['observation_time_information'] = lines_and_times
        fpos = fpos + int(header['block9']['blocklength'])
        self._check_fpos(fp_, fpos, 40, 'block9')
        fp_.seek(fpos, 0)

        header["block10"] = np.fromfile(fp_,
                                        dtype=_ERROR_INFO_TYPE,
                                        count=1)
        dtype = np.dtype([
            ("line_number", "<u2"),
            ("numof_error_pixels_per_line", "<u2"),
        ])
        num_err_info_data = header["block10"][
            'number_of_error_info_data'][0]
        err_info_data = []
        for _i in range(num_err_info_data):
            err_info_data.append(np.fromfile(fp_, dtype=dtype, count=1))
        header['error_information_data'] = err_info_data
        fpos = fpos + int(header['block10']['blocklength'])
        self._check_fpos(fp_, fpos, 40, 'block10')
        fp_.seek(fpos, 0)

        header["block11"] = np.fromfile(fp_, dtype=_SPARE_TYPE, count=1)
        fpos = fpos + int(header['block11']['blocklength'])
        self._check_fpos(fp_, fpos, 0, 'block11')
        fp_.seek(fpos, 0)

        return header

    def _read_data(self, fp_, header):
        """Read data block."""

        nlines = int(header["block2"]['number_of_lines'][0])
        ncols = int(header["block2"]['number_of_columns'][0])
        return da.from_array(np.memmap(self.filename, offset=fp_.tell(),
                                       dtype='<u2', shape=(nlines, ncols), mode='r'),
                             chunks=4096)

    def unzip_file(self, filename):
        """Unzip the file if file is bzipped = ending with 'bz2'."""
        # pathin = os.path.dirname(filename)
        # name = os.path.basename(filename)
        if filename.endswith('bz2'):
            # fdn, tmpfilepath = tempfile.mkstemp()
            tmpfilepath = filename.replace('.bz2','')
            if os.path.isfile(tmpfilepath) :
                return tmpfilepath
            print("Using temp file for BZ2 decompression: %s", tmpfilepath)
            # try pbzip2
            pbzip = which('pbzip2')
            # Run external pbzip2
            if pbzip is not None:
                n_thr = os.environ.get('OMP_NUM_THREADS')
                if n_thr:
                    runner = [pbzip,
                              '-dc',
                              '-p'+str(n_thr),
                              filename]
                else:
                    runner = [pbzip,
                              '-dc',
                              filename]
                p = Popen(runner, stdout=PIPE, stderr=PIPE)
                stdout = BytesIO(p.communicate()[0])
                status = p.returncode
                if status != 0:
                    raise IOError("pbzip2 error '%s', failed, status=%d"
                                  % (filename, status))
                with closing(open(tmpfilepath, 'wb')) as ofpt:
                    try:
                        stdout.seek(0)
                        shutil.copyfileobj(stdout, ofpt)
                    except IOError:
                        import traceback
                        traceback.print_exc()
                        print("Failed to read bzipped file %s",
                              str(filename))
                        os.remove(tmpfilepath)
                        raise
                return tmpfilepath

            # Otherwise, fall back to the original method
            bz2file = bz2.BZ2File(filename)
            with closing(open(tmpfilepath, 'wb')) as ofpt:
                try:
                    ofpt.write(bz2file.read())
                    ofpt.close()
                except IOError:
                    import traceback
                    traceback.print_exc()
                    print("Failed to read bzipped file %s", str(filename))
                    # os.remove(tmpfilepath)
                    return None
            return tmpfilepath

        return None

    def _check_fpos(self, fp_, fpos, offset, block):
        """Check file position matches blocksize."""
        if (fp_.tell() + offset != fpos):
            print("Actual "+block+" header size does not match expected")
        return

    def _mask_invalid(self, data, header):
        """Mask invalid data."""

        invalid = da.logical_or(data == header['block5']["count_value_outside_scan_pixels"][0],
                                data == header['block5']["count_value_error_pixels"][0])
        return da.where(invalid, np.float32(np.nan), data)

    def _mask_space(self, data):
        """Mask space pixels."""
        return data.where(get_geostationary_mask(self.area))
        # pass

    @property
    def start_time(self):
        """Get the start time."""
        return datetime(1858, 11, 17) + timedelta(days=float(self.basic_info['observation_start_time']))

    @property
    def end_time(self):
        """Get the end time."""
        return datetime(1858, 11, 17) + timedelta(days=float(self.basic_info['observation_end_time']))

    def get_area_def(self):
        """Get the area definition."""

        pdict = {}
        pdict['cfac'] = np.uint32(self.proj_info['CFAC'])
        pdict['lfac'] = np.uint32(self.proj_info['LFAC'])
        pdict['coff'] = np.float32(self.proj_info['COFF'])
        pdict['loff'] = -np.float32(self.proj_info['LOFF']) + 1
        pdict['a'] = float(self.proj_info['earth_equatorial_radius'] * 1000)
        pdict['h'] = float(self.proj_info['distance_from_earth_center'] * 1000 - pdict['a'])
        pdict['b'] = float(self.proj_info['earth_polar_radius'] * 1000)
        pdict['ssp_lon'] = float(self.proj_info['sub_lon'])
        pdict['nlines'] = int(self.data_info['number_of_lines'])
        pdict['ncols'] = int(self.data_info['number_of_columns'])
        pdict['scandir'] = 'N2S'

        pdict['loff'] = pdict['loff'] + (self.segment_number * pdict['nlines'])

        aex = get_area_extent(pdict)

        pdict['a_name'] = self.observation_area
        pdict['a_desc'] = "AHI {} area".format(self.observation_area)
        pdict['p_id'] = 'geosh8'

        area = get_area_definition(pdict, aex)

        self.area = area
        return area

    def latlon2ij(self, lat, lon):

        ea = float(self.proj_info['earth_equatorial_radius'])   # 6378.137          # 地球长半轴
        eb = float(self.proj_info['earth_polar_radius'])        # 6356.7523         # 地球短半轴
        H = float(self.proj_info['distance_from_earth_center'])    # 42164.0    # 地心到卫星质心的距离
        subpoint = float(self.proj_info['sub_lon'])
        coff = np.float32(self.proj_info['COFF'])
        cfac = np.uint32(self.proj_info['CFAC'])
        loff = np.float32(self.proj_info['LOFF'])
        lfac = np.uint32(self.proj_info['LFAC'])
        rowmax = int(self.data_info['number_of_columns'])
        colmax = int(self.data_info['number_of_columns'])

        lat1 = lat.copy()
        lon1 = lon.copy()
        deg2rad = np.pi / 180.0
        rad2deg = 180.0 / np.pi

        lon1[lon1 > 180] -= 360.0

        lon1 = lon1 * deg2rad
        lat1 = lat1 * deg2rad

        phi_e = np.arctan(eb ** 2 * np.tan(lat1) / ea ** 2)
        re = eb / np.sqrt(1-(ea**2 - eb**2) * np.cos(phi_e)**2/ea**2)
        r1 = H - re * np.cos(phi_e) * np.cos(lon1 - subpoint* deg2rad)
        r2 = -re * np.cos(phi_e) * np.sin(lon1 - subpoint* deg2rad)
        r3 = re * np.sin(phi_e)
        rn = np.sqrt(r1**2 + r2**2 + r3**2)
        x = np.arctan2(-r2, r1)*180/np.pi
        # x = np.arctan(-r2/r1)*180/np.pi
        y = np.arcsin(-r3/rn)*180/np.pi
        col = coff + x * 2**(-16) * cfac
        line = loff + y * 2**(-16) * lfac

        # 对外太空进行判断
        tmp = r1 * (r1 - H) + r2 **2 + r3**2
        flag = tmp > 0

        line = np.array(line+0.5, dtype=np.int32)
        col = np.array(col+0.5, dtype=np.int32)

        line[(line<0) | (line >= rowmax)] = -1
        col[(col<0) | (col>=colmax)]  = -1

        return line, col

    def __del__(self):

        # time.sleep(2)
        # if (self.is_zipped and os.path.exists(self.filename)):
        #     os.remove(self.filename)
        pass




def np2str(value):
    """Convert an `numpy.string_` to str.

    Args:
        value (ndarray): scalar or 1-element numpy array to convert

    Raises:
        ValueError: if value is array larger than 1-element or it is not of
                    type `numpy.string_` or it is not a numpy array

    """
    if hasattr(value, 'dtype') and \
            issubclass(value.dtype.type, (np.str_, np.string_, np.object_)) \
            and value.size == 1:
        value = value.item()
        if not isinstance(value, str):
            # python 3 - was scalar numpy array of bytes
            # otherwise python 2 - scalar numpy array of 'str'
            value = value.decode()
        return value
    else:
        raise ValueError("Array is not a string type or is larger than 1")

def get_user_calibration_factors(band_name, correction_dict):
    """Retrieve radiance correction factors from user-supplied dict."""
    if band_name in correction_dict:
        try:
            slope = correction_dict[band_name]['slope']
            offset = correction_dict[band_name]['offset']
        except KeyError:
            raise KeyError("Incorrect correction factor dictionary. You must "
                           "supply 'slope' and 'offset' keys.")
    else:
        # If coefficients not present, warn user and use slope=1, offset=0
        print("WARNING: You have selected radiance correction but "
                      " have not supplied coefficients for channel " +
                      band_name)
        return 1., 0.

    return slope, offset

def apply_rad_correction(data, slope, offset):
    """Apply GSICS-like correction factors to radiance data."""
    data = (data - offset) / slope
    return data

def get_area_extent(pdict):
    """Get the area extent seen by a geostationary satellite.

    Args:
        pdict: A dictionary containing common parameters:
            nlines: Number of lines in image
            ncols: Number of columns in image
            cfac: Column scaling factor
            lfac: Line scaling factor
            coff: Column offset factor
            loff: Line offset factor
            scandir: 'N2S' for standard (N->S), 'S2N' for inverse (S->N)
    Returns:
        aex: An area extent for the scene

    """
    # count starts at 1
    cols = 1 - 0.5

    if pdict['scandir'] == 'S2N':
        lines = 0.5 - 1
        scanmult = -1
    else:
        lines = 1 - 0.5
        scanmult = 1
    # Lower left x, y scanning angles in degrees
    ll_x, ll_y = get_xy_from_linecol(lines * scanmult,
                                     cols,
                                     (pdict['loff'], pdict['coff']),
                                     (pdict['lfac'], pdict['cfac']))

    cols += pdict['ncols']
    lines += pdict['nlines']
    # Upper right x, y scanning angles in degrees
    ur_x, ur_y = get_xy_from_linecol(lines * scanmult,
                                     cols,
                                     (pdict['loff'], pdict['coff']),
                                     (pdict['lfac'], pdict['cfac']))
    if pdict['scandir'] == 'S2N':
        ll_y *= -1
        ur_y *= -1

    # Convert degrees to radians and create area extent
    aex = make_ext(ll_x=ll_x, ur_x=ur_x, ll_y=ll_y, ur_y=ur_y, h=pdict['h'])

    return aex


def get_area_definition(pdict, a_ext):
    """Get the area definition for a geo-sat.

    Args:
        pdict: A dictionary containing common parameters:
            nlines: Number of lines in image
            ncols: Number of columns in image
            ssp_lon: Subsatellite point longitude (deg)
            a: Earth equatorial radius (m)
            b: Earth polar radius (m)
            h: Platform height (m)
            a_name: Area name
            a_desc: Area description
            p_id: Projection id
        a_ext: A four element tuple containing the area extent (scan angle)
               for the scene in radians
    Returns:
        a_def: An area definition for the scene

    .. note::

        The AreaDefinition `proj_id` attribute is being deprecated.

    """
    proj_dict = {'a': float(pdict['a']),
                 'b': float(pdict['b']),
                 'lon_0': float(pdict['ssp_lon']),
                 'h': float(pdict['h']),
                 'proj': 'geos',
                 'units': 'm'}

    a_def = AreaDefinition(
        pdict['a_name'],
        pdict['a_desc'],
        pdict['p_id'],
        proj_dict,
        int(pdict['ncols']),
        int(pdict['nlines']),
        a_ext)

    return a_def

def get_xy_from_linecol(line, col, offsets, factors):
    """Get the intermediate coordinates from line & col.

    Intermediate coordinates are actually the instruments scanning angles.
    """
    loff, coff = offsets
    lfac, cfac = factors
    x__ = float(col - coff) / (float(cfac) / 2 ** 16)
    y__ = float(line - loff) / (float(lfac) / 2 ** 16)

    return x__, y__


def make_ext(ll_x, ur_x, ll_y, ur_y, h):
    """Create the area extent from computed ll and ur.

    Args:
        ll_x: The lower left x coordinate (m)
        ur_x: The upper right x coordinate (m)
        ll_y: The lower left y coordinate (m)
        ur_y: The upper right y coordinate (m)
        h: The satellite altitude above the Earth's surface
    Returns:
        aex: An area extent for the scene

    """
    aex = (np.deg2rad(ll_x) * h, np.deg2rad(ll_y) * h,
           np.deg2rad(ur_x) * h, np.deg2rad(ur_y) * h)

    return aex


def get_geostationary_mask(area):
    """Compute a mask of the earth's shape as seen by a geostationary satellite.

    Args:
        area (pyresample.geometry.AreaDefinition) : Corresponding area
                                                    definition

    Returns:
        Boolean mask, True inside the earth's shape, False outside.

    """
    # Compute projection coordinates at the earth's limb
    h = area.proj_dict['h']
    xmax, ymax = get_geostationary_angle_extent(area)
    xmax *= h
    ymax *= h

    # Compute projection coordinates at the centre of each pixel
    x, y = area.get_proj_coords(chunks=4096)

    # Compute mask of the earth's elliptical shape
    return ((x / xmax) ** 2 + (y / ymax) ** 2) <= 1

def get_geostationary_angle_extent(geos_area):
    """Get the max earth (vs space) viewing angles in x and y."""
    # TODO: take into account sweep_axis_angle parameter

    # get some projection parameters
    try:
        crs = geos_area.crs
        a = crs.ellipsoid.semi_major_metre
        b = crs.ellipsoid.semi_minor_metre
        if np.isnan(b):
            # see https://github.com/pyproj4/pyproj/issues/457
            raise AttributeError("'semi_minor_metre' attribute is not valid "
                                 "in older versions of pyproj.")
    except AttributeError:
        # older versions of pyproj don't have CRS objects
        from pyresample.utils import proj4_radius_parameters
        a, b = proj4_radius_parameters(geos_area.proj_dict)

    req = float(a) / 1000
    rp = float(b) / 1000
    h = float(geos_area.proj_dict['h']) / 1000 + req

    # compute some constants
    aeq = 1 - req**2 / (h ** 2)
    ap_ = 1 - rp**2 / (h ** 2)

    # generate points around the north hemisphere in satellite projection
    # make it a bit smaller so that we stay inside the valid area
    xmax = np.arccos(np.sqrt(aeq))
    ymax = np.arccos(np.sqrt(ap_))
    return xmax, ymax



# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : basic.py

@Modify Time :  2022/10/31 15:52   

@Author      : Lee    

@Version     : 1.0   

@Description :

'''
import os
import sys
import numpy as np
import datetime

Deg2Rad = np.pi / 180.0
Rad2Deg = 180.0 / np.pi

def calc_earth_distance(lat1, lon1, lat2, lon2) :
    '''  计算球面两点之间距离（KM） '''
    lon11 = lon1 * Deg2Rad
    lat11 = lat1 * Deg2Rad
    lon21 = lon2 * Deg2Rad
    lat21 = lat2 * Deg2Rad

    dlon = lon21 - lon11
    dlat = lat21 - lat11
    h = np.sin(dlat/2)**2 + np.cos(lat11) * np.cos(lat21) * np.sin(dlon/2)**2
    distance = 2 * 6371.009 * np.arcsin(np.sqrt( h )) # 地球平均半径，6371km

    return distance

def planck_t2r(bt, wn):
    '''
    普朗克函数：将亮温（K）转成辐射值

    Parameters
    ----------
    bt : numpy.narray
        bright temperature, units: K
    wn : float or numpy.narray
        wave number(cm^-1)

    Returns
    -------
        numpy.narray
        卫星观测辐射值，mW/(m2.cm-1.sr)

    '''
    Radiation_C1 = 1.191042953E-5
    Radiation_C2 = 1.4387774
    a = Radiation_C1 * wn * wn * wn
    b = (Radiation_C2 * wn) / bt
    c = np.exp(b) - 1.0
    radi = a / c

    return radi


def planck_r2t(rad, wn):
    '''
    普朗克函数：将辐射值转成亮温（K）

    Parameters
    ----------
    rad : numpy.narray
        mW/(m2.cm-1.sr)
    wn : float or numpy.narray
        wave number(cm^-1)

    Returns
    -------
        numpy.narray
        bright temperature
        units: K
    '''

    Radiation_C1 = 1.191042953E-5
    Radiation_C2 = 1.4387774
    vs = wn
    bt = (Radiation_C2*vs/np.log(Radiation_C1*vs*vs*vs/(rad)+1.0))

    return bt


def k_index(pressure, temperature, dewpoint):
    '''
    Calculate K Index from the pressure temperature and dewpoint.

    K Index formula derived from [George1960]:
        K = (T850 - T500) + Td850 - (T700 - Td700)

    where:
    T850 is the temperature at 850 hPa
    T700 is the temperature at 700 hPa
    T500 is the temperature at 500 hPa
    Td850 is the dewpoint at 850 hPa
    Td700 is the dewpoint at 700 hPa

    Calculation of the K Index is defined as the temperature
    difference between the static instability between 850 hPa and 500 hPa,
     add with the moisture at 850hPa, then subtract from the dryness of the airmass at 700 hPa.

    Parameters
    ----------
    pressure
    temperature
    dewpoint

    Returns
    -------

    '''

def total_totals_index(pressure, temperature, dewpoint):
    '''
    Calculate Total Totals Index from the pressure temperature and dewpoint.

    Total Totals Index formula derived from [Miller1972]:
        TT = (T850 + Td850) - (2 * T500)

    where:

    T850 is the temperature at 850 hPa
    T500 is the temperature at 500 hPa
    Td850 is the dewpoint at 850 hPa

    Calculation of the Total Totals Index is defined as
    the temperature at 850 hPa plus the dewpoint at 850 hPa,
    minus twice the temperature at 500 hPa.
    This index consists of two components,
    the Vertical Totals (VT) and the Cross Totals (CT).

    Parameters
    ----------
    pressure
    temperature
    dewpoint

    Returns
    -------

    '''

def lifted_index(pressure, temperature, parcel_profile):
    '''
    Calculate Lifted Index from the pressure temperature and parcel profile.

    Lifted index formula derived from [Galway1956] and referenced by [DoswellSchultz2006]:
        LI = T500 - Tp500

    where:
    T500 is the measured temperature at 500 hPa
     Tp500 is the temperature of the lifted parcel at 500 hPa

    Calculation of the lifted index is defined as the temperature
     difference between the observed 500 hPa temperature and the
     temperature of a parcel lifted from the surface to 500 hPa.

    Parameters
    ----------
    pressure
    temperature
    parcel_profile

    Returns
    -------

    '''


def wind_direction(u, v, convention='from'):
    '''
    Compute the wind direction from u and v-components.

    Parameters
    ----------
    u
    v
    convention

    Returns
    -------

    '''



# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : AtmCorr_GF.py

@Modify Time :  2022/12/23 16:38   

@Author      : Lee    

@Version     : 1.0   

@Description :

'''
import os
import shutil
import sys
import numpy as np
import datetime
import json
from tqdm import tqdm
from .AtmCorr import AtmCorr
from osgeo import gdal, gdalconst
from dateutil.relativedelta import relativedelta
import xml.dom.minidom
EXEPATH = os.path.split(os.path.realpath(__file__))[0]

class AtmCorr_GF(AtmCorr):

    def __init__(self):
        super(AtmCorr_GF, self).__init__()

    def FLAASH(self, outname, tiffFile, metedata, nowdate, SatID, InstID,
               InstType=None, fillvalue=65535, blocksize=1000):

        outdir = os.path.dirname(outname)
        if not os.path.isdir(outdir) :
            os.makedirs(outdir)
            print('成功创建路径【%s】' %(outdir))

        if not os.path.isfile(tiffFile) :
            print('文件不存在【%s】' %(tiffFile))
            return None

        srcdataset = gdal.Open(tiffFile, gdal.GA_ReadOnly)
        if srcdataset == None:
            print("文件无法打开【%s】" %(tiffFile))
            return None

        cols = srcdataset.RasterXSize                  # 栅格矩阵的列数
        rows = srcdataset.RasterYSize                  # 栅格矩阵的行数
        Bands = srcdataset.RasterCount                 # 波段数

        geoTransform1 = srcdataset.GetGeoTransform()   # 获取仿射矩阵信息
        proj1 = srcdataset.GetProjection()             # 获取投影信息

        # 创建输出结果文件
        driver = gdal.GetDriverByName("GTiff")
        dstdataset = driver.Create(outname, cols, rows, Bands, gdal.GDT_UInt16,
                                   options=["COMPRESS=LZW", "BigTIFF=YES"])
        dstdataset.SetGeoTransform(geoTransform1)
        dstdataset.SetProjection(proj1)

        #分别读取4个波段
        for BandID in range(0, Bands):
            ReadBand = srcdataset.GetRasterBand(BandID+1)
            outband = dstdataset.GetRasterBand(BandID+1)
            outband.SetNoDataValue(fillvalue)
            #获取对应波段的增益gain和偏移bias
            Gain, Bias = self.RadiometricCalibration(nowdate, SatID, InstID)

            if 'PMS' in InstID :
                if InstType is None :
                    raise Exception('请传入InstType参数【全色：PAN，多光谱：MSS】')

                if 'MSS' in InstType :
                    BandID += 1
                elif 'PAN' in InstType :
                    BandID = BandID
                else:
                    raise Exception('请传入InstType参数【全色：PAN，多光谱：MSS】')

            #获取大气校正系数
            # if 'PMS' in InstID :
            #     if 'PAN' not in InstType :
            #         self.setParam(nowdate, metedata, SatID, InstID, BandID+1)
            #         AtcCofa, AtcCofb, AtcCofc = self.corrCoeff()

            self.setParam(nowdate, metedata, SatID, InstID, BandID+1)
            AtcCofa, AtcCofb, AtcCofc = self.corrCoeff()


            i = 0
            j = 0
            #进度条参数
            XBlockcount = np.ceil(cols / blocksize)
            YBlockcount = np.ceil(rows / blocksize)
            # print("第%d波段校正："%BandID)
            try:
                with tqdm(total=XBlockcount*YBlockcount, iterable='iterable',
                          desc = '正在进行第%i波段校正' %(BandID+1), mininterval=1) as pbar:
                    while i < rows:
                        while j < cols:
                            # 保存分块大小
                            nXBK = blocksize
                            nYBK = blocksize

                            # 最后一块
                            if i+blocksize>rows:
                                nYBK = rows - i
                            if j+blocksize>cols:
                                nXBK=cols - j

                            # 分块读取影像
                            Image = ReadBand.ReadAsArray(j, i, nXBK,nYBK)

                            # if 'PMS' in InstID :
                            #     if 'PAN' in InstType :  # 全色波段将不做大气校正，直接返回
                            #         outImage =np.where(Image>0, (Image*Gain[BandID] + Bias[BandID])*10000, fillvalue)
                            #         outband.WriteArray(outImage,j,i)
                            #         j=j+nXBK
                            #         pbar.update(1)
                            #         continue

                            outImage =np.where(Image>0, Image*Gain[BandID] + Bias[BandID], fillvalue)

                            y = np.where(outImage!=fillvalue, AtcCofa * outImage - AtcCofb, fillvalue)
                            atcImage = np.where(y!=fillvalue, (y / (1 + y * AtcCofc))*10000, fillvalue)

                            outband.WriteArray(atcImage,j,i)
                            j=j+nXBK
                            pbar.update(1)
                        j=0
                        i=i+nYBK
            except BaseException :
                pbar.close()
                continue
            pbar.close()

        dstdataset = None
        srcdataset = None

    def setParam(self, nowdate, metedata, SatID, InstID, BandId, dem=0.010):

        # 读取头文件
        dom = xml.dom.minidom.parse(metedata)

        # 太阳和卫星角度信息
        sunz = 90 - float(dom.getElementsByTagName('SolarZenith')[0].firstChild.data)
        suna = float(dom.getElementsByTagName('SolarAzimuth')[0].firstChild.data)
        # s.geometry.view_z = float(dom.getElementsByTagName('SatelliteZenith')[0].firstChild.data)
        # s.geometry.view_a = float(dom.getElementsByTagName('SatelliteAzimuth')[0].firstChild.data)
        satz = 0
        sata = 0
        # 日期
        DateTimeparm = dom.getElementsByTagName('CenterTime')[0].firstChild.data
        DateTime = datetime.datetime.strptime(DateTimeparm, '%Y-%m-%d %H:%M:%S')

        # 中心经纬度
        TopLeftLat     = float(dom.getElementsByTagName('TopLeftLatitude')[0].firstChild.data)
        TopLeftLon     = float(dom.getElementsByTagName('TopLeftLongitude')[0].firstChild.data)
        TopRightLat    = float(dom.getElementsByTagName('TopRightLatitude')[0].firstChild.data)
        TopRightLon    = float(dom.getElementsByTagName('TopRightLongitude')[0].firstChild.data)
        BottomRightLat = float(dom.getElementsByTagName('BottomRightLatitude')[0].firstChild.data)
        BottomRightLon = float(dom.getElementsByTagName('BottomRightLongitude')[0].firstChild.data)
        BottomLeftLat  = float(dom.getElementsByTagName('BottomLeftLatitude')[0].firstChild.data)
        BottomLeftLon  = float(dom.getElementsByTagName('BottomLeftLongitude')[0].firstChild.data)

        ImageCenterLat = (TopLeftLat + TopRightLat + BottomRightLat + BottomLeftLat) / 4

        self.set_geom(nowdate, sunz=sunz, suna=suna, satz=satz, sata=sata)
        self.set_atm(sLatitude=ImageCenterLat, nowdate=nowdate)
        self.set_aer()
        self.set_vis()
        self.set_altitude(dem=dem)

        #读取辐射校正和大气校正所需参数:增益、偏移和光谱响应函数
        SRFFile = os.path.join(EXEPATH, 'resp', 'GF', "GF.json")
        if not os.path.isfile(SRFFile) :
            raise Exception('高分卫星光谱响应文件不存在【%s】' %(SRFFile))
        SRF = json.load(open(SRFFile))

        minwl = SRF[SatID][InstID]['B%d' %(BandId)]['wl'][0]
        maxwl = SRF[SatID][InstID]['B%d' %(BandId)]['wl'][1]
        response = SRF[SatID][InstID]['B%d' %(BandId)]['SRF']
        self.set_resp(minwl, maxwl, response)

    def RadiometricCalibration(self, nowdate, SatID, InstID):
        ''' 获取高分卫星定标系数 gain和 bias '''

        #读取辐射校正和大气校正所需参数:增益、偏移和光谱响应函数
        CalCoeffFile = os.path.join(EXEPATH, "resp", "GF", "CalibrationCoefficient.json")
        if not os.path.isfile(CalCoeffFile) :
            raise Exception('高分卫星定标系数文件不存在【%s】' %(CalCoeffFile))
        CalCoeff = json.load(open(CalCoeffFile))

        satid = SatID.replace('-', '')
        instid = InstID.replace('-', '')

        if not satid in CalCoeff :
            raise Exception('请确认【%s】是否在定标系数列表，当前仅支持' %(SatID), CalCoeff.keys())

        if not instid in CalCoeff[satid] :
            raise Exception('请确认【%s】是否在定标系数列表，当前仅支持' %(InstID), CalCoeff[satid].keys())

        while nowdate >= datetime.datetime.strptime('2000', '%Y') :
            stryear = nowdate.strftime('%Y')
            if not stryear in CalCoeff[satid][instid] :
                nowdate -= relativedelta(years=1)
            else:
                gain = CalCoeff[satid][instid][stryear]['gain']
                bias = CalCoeff[satid][instid][stryear]['bias']
                break

        return gain, bias

    def ortho_rectification(self, dstfile, srcfile, rpcfile, demfile=None):
        """
        正射校正
        :param input:输入原始影像
        :param output:输出正射影像
        """

        if not os.path.isfile(rpcfile) :
            raise Exception('RPC文件不存在【%s】' %(rpcfile))

        srcdir = os.path.dirname(srcfile)
        shutil.copy(rpcfile, srcdir)

        dataset = gdal.Open(srcfile, gdal.GA_Update)#读入影像
        rpc = dataset.GetMetadata("RPC")#读入影像，rpc

        if demfile is None :
            dst_ds = gdal.Warp(dstfile, dataset, dstSRS='EPSG:4326',
                               # xRes=resolution,
                               # yRes=resolution,
                               resampleAlg=gdal.GRIORA_Bilinear,
                               rpc=True, #使用RPC模型进行校正
                               warpOptions=['INIT_DEST=NO_DATA'],
                               creationOptions=["COMPRESS=LZW", "BigTIFF=YES"])
        else:
            dst_ds = gdal.Warp(dstfile, dataset, dstSRS='EPSG:4326',
                               # xRes=resolution,
                               # yRes=resolution,
                               # resampleAlg=gdal.GRIORA_Bilinear,
                               rpc=True, #使用RPC模型进行校正
                               transformerOptions=[r'RPC_DEM=%s' %(demfile),
                                                   "RPC_DEMINTERPOLATION=bilinear"],
                               creationOptions=["COMPRESS=LZW", "BigTIFF=YES"])

        dst_ds = None

    def pansharpen(self,
                   dst_filename,
                   pan_name,
                   dict_spectrac=None,
                   format = None,
                   resampleAlg= None,
                   nodata_value = None,
                   weights = None,
                   creation_options = None,
                   spat_adjust = None,
                   num_threads = None,
                   bitdepth = None,
                   verbose_vrt = False,
                   progress_callback = gdal.TermProgress_nocb):

        band_nums = []
        weights = weights or []
        creation_options = creation_options or []
        spectral_ds = None
        spectral_bands = None

        if dict_spectrac :
            spectral_ds, spectral_bands = self.parse_spectral_names(dict_spectrac)

        if pan_name is None or spectral_bands is None:
            return 1

        # 打开PAN图像
        pan_ds = gdal.Open(pan_name)
        if pan_ds is None:
            return 1

        if format is None:
            format = self.GetOutputDriverFor(dst_filename)

        if not band_nums:
            band_nums = [j + 1 for j in range(len(spectral_bands))]
        else:
            for band in band_nums:
                if band < 0 or band > len(spectral_bands):
                    print('Invalid band number in -b: %d' % band)
                    return 1

        if weights and len(weights) != len(spectral_bands):
            print('There must be as many -w values specified as input spectral bands')
            return 1

        # 开始构建VRT文件
        vrt_xml = """<VRTDataset subClass="VRTPansharpenedDataset">\n"""
        if band_nums != [j + 1 for j in range(len(spectral_bands))]:
            for i, band in enumerate(band_nums):
                sband = spectral_bands[band - 1]
                datatype = gdal.GetDataTypeName(sband.DataType)
                colorname = gdal.GetColorInterpretationName(sband.GetColorInterpretation())
                vrt_xml += """  <VRTRasterBand dataType="%s" band="%d" subClass="VRTPansharpenedRasterBand">
          <ColorInterp>%s</ColorInterp>
      </VRTRasterBand>\n""" % (datatype, i + 1, colorname)

        vrt_xml += """  <PansharpeningOptions>\n"""

        # 对各个波段增加权重
        if weights:
            vrt_xml += """      <AlgorithmOptions>\n"""
            vrt_xml += """        <Weights>"""
            for i, weight in enumerate(weights):
                if i > 0:
                    vrt_xml += ","
                vrt_xml += "%.16g" % weight
            vrt_xml += "</Weights>\n"
            vrt_xml += """      </AlgorithmOptions>\n"""

        if resampleAlg is not None:
            vrt_xml += f'      <Resampling>{resampleAlg}</Resampling>\n'

        if num_threads is not None:
            vrt_xml += f'      <NumThreads>{num_threads}</NumThreads>\n'

        if bitdepth is not None:
            vrt_xml += f'      <BitDepth>{bitdepth}</BitDepth>\n'

        if nodata_value is not None:
            vrt_xml += f'      <NoData>{nodata_value}</NoData>\n'

        if spat_adjust is not None:
            vrt_xml += f'      <SpatialExtentAdjustment>{spat_adjust}</SpatialExtentAdjustment>\n'

        pan_relative = '0'
        if format.upper() == 'VRT':
            if not os.path.isabs(pan_name):
                pan_relative = '1'
                pan_name = os.path.relpath(pan_name, os.path.dirname(dst_filename))

        vrt_xml += """    <PanchroBand>
          <SourceFilename relativeToVRT="%s">%s</SourceFilename>
          <SourceBand>1</SourceBand>
        </PanchroBand>\n""" % (pan_relative, pan_name)

        for i, sband in enumerate(spectral_bands):
            dstband = ''
            for j, band in enumerate(band_nums):
                if i + 1 == band:
                    dstband = ' dstBand="%d"' % (j + 1)
                    break

            ms_relative = '0'
            ms_name = spectral_ds[i].GetDescription()
            if format.upper() == 'VRT':
                if not os.path.isabs(ms_name):
                    ms_relative = '1'
                    ms_name = os.path.relpath(ms_name, os.path.dirname(dst_filename))

            vrt_xml += """    <SpectralBand%s>
          <SourceFilename relativeToVRT="%s">%s</SourceFilename>
          <SourceBand>%d</SourceBand>
        </SpectralBand>\n""" % (dstband, ms_relative, ms_name, sband.GetBand())

        vrt_xml += """  </PansharpeningOptions>\n"""
        vrt_xml += """</VRTDataset>\n"""

        if format.upper() == 'VRT':
            f = gdal.VSIFOpenL(dst_filename, 'wb')
            if f is None:
                print('Cannot create %s' % dst_filename)
                return 1
            gdal.VSIFWriteL(vrt_xml, 1, len(vrt_xml), f)
            gdal.VSIFCloseL(f)
            if verbose_vrt:
                vrt_ds = gdal.Open(dst_filename, gdal.GA_Update)
                vrt_ds.SetMetadata(vrt_ds.GetMetadata())
            else:
                vrt_ds = gdal.Open(dst_filename)
            if vrt_ds is None:
                return 1

            return 0

        vrt_ds = gdal.Open(vrt_xml)
        out_ds = gdal.GetDriverByName(format).CreateCopy(dst_filename, vrt_ds, 0, creation_options,
                                                         callback=progress_callback)
        if out_ds is None:
            return 1
        return 0

    def DoesDriverHandleExtension(self, drv, ext):
        exts = drv.GetMetadataItem(gdal.DMD_EXTENSIONS)
        return exts is not None and exts.lower().find(ext.lower()) >= 0

    def GetExtension(self, filename):
        ext = os.path.splitext(filename)[1]
        if ext.startswith('.'):
            ext = ext[1:]
        return ext


    def GetOutputDriversFor(self, filename):
        drv_list = []
        ext = self.GetExtension(filename)
        for i in range(gdal.GetDriverCount()):
            drv = gdal.GetDriver(i)
            if (drv.GetMetadataItem(gdal.DCAP_CREATE) is not None or
                drv.GetMetadataItem(gdal.DCAP_CREATECOPY) is not None) and \
                    drv.GetMetadataItem(gdal.DCAP_RASTER) is not None:
                if ext and self.DoesDriverHandleExtension(drv, ext):
                    drv_list.append(drv.ShortName)
                else:
                    prefix = drv.GetMetadataItem(gdal.DMD_CONNECTION_PREFIX)
                    if prefix is not None and filename.lower().startswith(prefix.lower()):
                        drv_list.append(drv.ShortName)

        # GMT is registered before netCDF for opening reasons, but we want
        # netCDF to be used by default for output.
        if ext.lower() == 'nc' and not drv_list and \
                drv_list[0].upper() == 'GMT' and drv_list[1].upper() == 'NETCDF':
            drv_list = ['NETCDF', 'GMT']

        return drv_list


    def GetOutputDriverFor(self,
                           filename=None,
                           is_raster=True,
                           default_raster_format='GTiff',
                           default_vector_format='ESRI Shapefile'):
        if not filename:
            return 'MEM'
        drv_list = self.GetOutputDriversFor(filename)
        ext = self.GetExtension(filename)
        if not drv_list:
            if not ext:
                return default_raster_format if is_raster else default_vector_format
            else:
                raise Exception("Cannot guess driver for %s" % filename)
        elif len(drv_list) > 1:
            print("Several drivers matching %s extension. Using %s" % (ext if ext else '', drv_list[0]))
        return drv_list[0]

    def parse_spectral_names(self, dict_spectrac):
        spectral_ds = []
        spectral_bands = []

        for spectral_name in dict_spectrac :
            ds = gdal.Open(spectral_name)
            if ds is None:
                print('读取文件失败【%s】' %(spectral_name))
                return None,None

            # 根据指定波段获取dataset ID
            for band_num in dict_spectrac[spectral_name] :
                band = ds.GetRasterBand(band_num)
                spectral_ds.append(ds)
                spectral_bands.append(band)

        return spectral_ds, spectral_bands



# -*- coding:utf-8 -*-
'''
@Project  : fypy

@File     : h4toh5.py

@Modify Time : 2022/8/11 15:34

@Author : Lee

@Version : 1.0

@Description :
将hdf4格式文件转换成hdf5文件

'''
import os
import sys
import h5py



def h4toh5(h4file, h5file):
    '''
    将hdf4格式文件转换成hdf5文件

    Parameters
    ----------
    h4file: str
        输入hdf4文件名
    h5file: str
        输出hdf5文件名
    Returns
    -------
        bool
        成功返回True, 失败返回False
    '''
    from pyhdf import SD

    if not os.path.isfile(h4file):
        print('%s is not exist, will continue...' %(h4file))
        return False

    fp4 = SD.SD(h4file, SD.SDC.READ)
    fp5 = h5py.File(h5file, 'w')
    attrs = fp4.attributes(full=1)

    # 将hdf4中的文件属性写入hdf5
    for item in fp4.attributes().keys():
        # print(item)
        try:
            fp5.attrs[item] = attrs[item][0]
        except BaseException as e:
            print(e)

    # 将hdf4中的数据集和属性写入hdf5
    for sdsname in fp4.datasets().keys() :
        # print(sdsname)
        try:
            sds4id = fp4.select(sdsname)
            data = sds4id[:]
            dsetid = fp5.create_dataset(name= sdsname, data=data, compression=9)
            attrs = sds4id.attributes(full=1)
            for key in sds4id.attributes() :
                dsetid.attrs[key] =  attrs[key][0]
        except BaseException as e:
            print(e)
    fp4.end()
    fp5.close()

    return True














# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : AtmCorr_Sentinel.py

@Modify Time :  2023/4/27 14:47   

@Author      : Lee    

@Version     : 1.0   

@Description :

'''
import os
import sys
import numpy as np
import datetime
import json
from tqdm import tqdm
from osgeo import gdal, osr, ogr
from dateutil.relativedelta import relativedelta
import xml.dom.minidom
from Py6S import *

from .AtmCorr import AtmCorr, reflectance2radiance

EXEPATH = os.path.split(os.path.realpath(__file__))[0]

class AtmCorr_Sentinel(AtmCorr):

    def __init__(self):
        super(AtmCorr_Sentinel, self).__init__()

    def FLAASH(self, outname, tiffFile, metadata, SatID, InstID, BandId,
               fillvalue=65535, blocksize=1000):

        outdir = os.path.dirname(outname)
        if not os.path.isdir(outdir) :
            os.makedirs(outdir)
            print('成功创建路径【%s】' %(outdir))

        if not os.path.isfile(tiffFile) :
            print('文件不存在【%s】' %(tiffFile))
            return None

        # 捕捉打开数据出错异常
        try:
            srcdataset = gdal.Open(tiffFile)
        except Exception as e:
            print("文件{file}打开失败".format(file=tiffFile))

        if srcdataset == None:
            print("{file}数据集读取为空".format(file=tiffFile))
            return None

        # 获取行列号
        cols = srcdataset.RasterXSize                  # 栅格矩阵的列数
        rows = srcdataset.RasterYSize                  # 栅格矩阵的行数
        Bands = srcdataset.RasterCount                 # 波段数

        srcTrans = srcdataset.GetGeoTransform()   # 获取仿射矩阵信息
        srcProj = srcdataset.GetProjection()             # 获取投影信息

        # 创建输出结果文件
        driver = gdal.GetDriverByName("GTiff")
        dstdataset = driver.Create(outname, cols, rows, Bands, gdal.GDT_UInt16,
                                   options=["COMPRESS=LZW", "BigTIFF=YES"])
        dstdataset.SetGeoTransform(srcTrans)
        dstdataset.SetProjection(srcProj)

        # 表观反射率转换为辐射亮度值
        CalCoeffFile = os.path.join(EXEPATH, "resp", "Sentinel",
                                    "CalibrationCoefficient.json")
        if not os.path.isfile(CalCoeffFile) :
            raise Exception('哨兵卫星定标系数文件不存在【%s】' %(CalCoeffFile))
        CalCoeff = json.load(open(CalCoeffFile))

        satid = SatID.replace('-', '')
        instid = InstID.replace('-', '')

        if not satid in CalCoeff :
            raise Exception('请确认【%s】是否在定标系数列表，当前仅支持' %(SatID), CalCoeff.keys())

        if not instid in CalCoeff[satid] :
            raise Exception('请确认【%s】是否在定标系数列表，当前仅支持' %(InstID), CalCoeff[satid].keys())

        Esun = CalCoeff[SatID][InstID]['ESUN'][BandId-1]

        #大气校正
        self.setParam(metadata['Date'], metadata, SatID, InstID, BandId, dem=0)
        Coeffa, Coeffb, Coeffc  = self.corrCoeff()

        # 创建输出结果文件
        driver = gdal.GetDriverByName("GTiff")
        dstdataset = driver.Create(outname, cols, rows, Bands, gdal.GDT_UInt16,
                                   options=["COMPRESS=LZW", "BigTIFF=YES"])
        dstdataset.SetGeoTransform(srcTrans)
        dstdataset.SetProjection(srcProj)

        ReadBand = srcdataset.GetRasterBand(1)
        outband = dstdataset.GetRasterBand(1)
        outband.SetNoDataValue(fillvalue)

        #进度条参数
        XBlockcount = np.ceil(cols / blocksize)
        YBlockcount = np.ceil(rows / blocksize)
        i = 0
        j = 0
        try:
            with tqdm(total=XBlockcount*YBlockcount, iterable='iterable',
                      desc = '正在进行第%i波段校正' %(BandId), mininterval=1) as pbar:
                while i < rows:
                    while j < cols:
                        # 保存分块大小
                        nXBK = blocksize
                        nYBK = blocksize

                        # 最后一块
                        if i+blocksize > rows:
                            nYBK = rows - i
                        if j+blocksize > cols:
                            nXBK=cols - j

                        # 分块读取影像
                        srcdata = ReadBand.ReadAsArray(j, i, nXBK,nYBK)
                        TOARef = srcdata * 0.0001
                        radiance = reflectance2radiance(TOARef, Esun, metadata['sunz'], metadata['Date'])

                        outImage =np.where(radiance>0, radiance, fillvalue)

                        y = np.where(outImage!=fillvalue, Coeffa * outImage - Coeffb, fillvalue)
                        atcImage = np.where(y!=fillvalue, (y / (1 + y * Coeffc))*10000, fillvalue)

                        outband.WriteArray(atcImage, j, i)
                        j=j+nXBK
                        pbar.update(1)
                    j=0
                    i=i+nYBK
        except BaseException :
            pbar.close()

        pbar.close()

    def setParam(self, nowdate, metadata, SatID, InstID, BandId, dem=0.010):
        '''

        :param nowdate:
        :param BandId:
        :param radiance:
        :param metadata:
        :param dem:
        :return:
        '''

        self.set_geom(nowdate, sunz=metadata['sunz'], suna=metadata['suna'], satz=0, sata=0)
        self.set_atm(sLatitude=metadata['centre_pos'][1], nowdate=nowdate)
        self.set_aer()
        self.set_vis()
        self.set_altitude(dem=dem)

        #读取辐射校正和大气校正所需参数:增益、偏移和光谱响应函数
        SRFFile = os.path.join(EXEPATH, 'resp', 'Sentinel', "Sentinel.json")
        if not os.path.isfile(SRFFile) :
            raise Exception('哨兵卫星光谱响应文件不存在【%s】' %(SRFFile))
        SRF = json.load(open(SRFFile))

        minwl = SRF[SatID][InstID]['B%d' %(BandId)]['wl'][0]
        maxwl = SRF[SatID][InstID]['B%d' %(BandId)]['wl'][1]
        response = SRF[SatID][InstID]['B%d' %(BandId)]['SRF']
        self.set_resp(minwl, maxwl, response)

        # self.set_resp(wavelength=wavelength)

    def GetMeta(self, metafile):
        dom = xml.dom.minidom.parse(metafile)
        dict_meta = {}

        #太阳天顶角、方位角
        SunAngle = dom.getElementsByTagName('Mean_Sun_Angle')
        dict_meta['sunz'] = float(SunAngle[0].getElementsByTagName('ZENITH_ANGLE')[0].firstChild.data)
        dict_meta['suna'] = float(SunAngle[0].getElementsByTagName('AZIMUTH_ANGLE')[0].firstChild.data)

        #卫星天顶角、方位角
        ViewAngles = dom.getElementsByTagName('Mean_Viewing_Incidence_Angle')
        dict_satz = {}
        dict_sata = {}

        for angle in ViewAngles:
            ViewAngle = int(angle.getAttribute('bandId'))
            dict_satz[ViewAngle+1] = float(angle.getElementsByTagName('ZENITH_ANGLE')[0].firstChild.data)
            dict_sata[ViewAngle+1]= float(angle.getElementsByTagName('AZIMUTH_ANGLE')[0].firstChild.data)

        dict_meta["ViewZenithAngle"] = dict_satz
        dict_meta["ViewAzimuthAngle"] = dict_sata
        # 日期:月、日  2023-04-11T03:50:21.042436Z
        dict_meta["Date"] = datetime.datetime.strptime(dom.getElementsByTagName('SENSING_TIME')[0].firstChild.data,
                                                       '%Y-%m-%dT%H:%M:%S.%fZ')

        #求影像中心经纬度
        PointULX = int(dom.getElementsByTagName('ULX')[0].firstChild.data)
        PointULY = int(dom.getElementsByTagName('ULY')[0].firstChild.data)

        Imgsizes = dom.getElementsByTagName('Size')

        for Imgsize in Imgsizes:
            Resolution = Imgsize.getAttribute('resolution')
            if Resolution == '10':
                dict_meta["Nrows"] = int(Imgsize.getElementsByTagName('NROWS')[0].firstChild.data)
                dict_meta["Ncols"] = int(Imgsize.getElementsByTagName('NCOLS')[0].firstChild.data)

        PointBRX = PointULX + 10*dict_meta["Ncols"]
        PointBRY = PointULY - 10*dict_meta["Nrows"]

        # 将投影坐标转为经纬度坐标（具体的投影坐标系由给定数据确定）
        Proj = dom.getElementsByTagName('HORIZONTAL_CS_CODE')[0].firstChild.data
        ProjCode = int(Proj.split(':')[1])

        source = osr.SpatialReference()
        source.ImportFromEPSG(ProjCode)
        target = osr.SpatialReference()
        target.ImportFromEPSG(4326)
        ct = osr.CoordinateTransformation(source,target)
        CoordsUL,CoordsBR = ct.TransformPoints([(PointULX,PointULY),(PointBRX,PointBRY)])

        ULLat = CoordsUL[0]
        ULLon = CoordsUL[1]
        BRLat = CoordsBR[0]
        BRLon = CoordsBR[1]

        sLongitude = (ULLon+BRLon) / 2
        sLatitude = (ULLat+BRLat) / 2

        dict_meta['extent'] = [ULLon, BRLat, BRLon, ULLat]  # minX, minY, maxX, maxY
        dict_meta['centre_pos'] = [sLongitude, sLatitude]

        return dict_meta


# -*- coding:utf-8 -*-
'''
@Project  : fypy

@File     : ncpro.py

@Modify Time : 2022/8/11 15:34

@Author : Lee

@Version : 1.0

@Description :
Netcdf4文件的读写操作

'''
import os
import netCDF4
import numpy as np
from netCDF4 import num2date, date2num

def readnc(filename, sdsname, dictsdsinfo=None):
    ''' 读取netcdf4 文件数据集和数据集属性 '''
    if not os.path.isfile(filename) :
        print('%s is not exist, please check it...' %(filename))
        if dictsdsinfo is None :
            return None
        else:
            return None, None
    else:
        try:
            with netCDF4.Dataset(filename, 'r') as fp :
                if not sdsname in fp.variables:
                    print('%s is not a dataset in %s' % (sdsname, filename))

                    if dictsdsinfo is None :
                        return None
                    else:
                        return None, None
                else:
                    dsetid = fp.variables[sdsname]
                    data = fp.variables[sdsname][:]

                    if not dictsdsinfo is None :
                        for key in dsetid.ncattrs() :
                            dictsdsinfo.update({key: dsetid.getncattr(key)})


                        return data, dictsdsinfo

        except BaseException as e :
            print(e)
            if dictsdsinfo is None :
                return None
            else:
                return None, None

        return data

def readnc_fileinfo(filename) :
    ''' 读取netcdf4 文件全局属性 '''
    dictfileinfo = {}
    if not os.path.isfile(filename) :
        print('%s is not exist, please check it...' %(filename))
        return dictfileinfo
    else:
        try:
            with netCDF4.Dataset(filename, 'r') as fp :

                for key in fp.ncattrs():
                    dictfileinfo.update({key: fp.getncattr(key)})

                return  dictfileinfo
        except BaseException as e :
            print(e)
            return dictfileinfo

def readnc_sdsinfo(filename, sdsname) :
    ''' 读取数据集属性信息 '''
    dictsdsinfo = {}
    if not os.path.isfile(filename) :
        print('%s is not exist, please check it...' %(filename))
        return dictsdsinfo
    else:
        try:
            with netCDF4.Dataset(filename, 'r') as fp :

                if not sdsname in fp.variables:
                    print('%s is not a dataset in %s' %(sdsname, filename))
                    return dictsdsinfo
                else:
                    dsetid = fp.variables[sdsname]
                    for key in dsetid.ncattrs():
                        dictsdsinfo.update({key: dsetid.getncattr(key)})

                    return dictsdsinfo
        except BaseException as e :
            print(e)
            return dictsdsinfo


def readncfortimes(filename, sdsname,
                   units = 'hours since 1900-01-01 00:00:00.0',
                   calendar = "gregorian"):
    '''
    读取时间维度数据集

    Parameters
    ----------
    filename : str
        文件名
    sdsname : str
        数据集名
    units : str
        时间单位，从什么时间开始计数
    calendar : str

    Returns
    -------
        numpy.array
        时间维度数据集
    '''

    if not os.path.isfile(filename) :
        print('%s is not exist, please check it...' %(filename))
        return None
    else:
        try:
            with netCDF4.Dataset(filename, 'r') as fp :
                if not sdsname in fp.variables:
                    print('%s is not a dataset in %s' % (sdsname, filename))
                    return None
                else:
                    dsetid = fp.variables[sdsname]
                    data = fp.variables[sdsname][:]

                    if 'units' in dsetid.ncattrs() :
                        units = dsetid.getncattr('units')

                    if 'calendar' in dsetid.ncattrs() :
                        calendar = dsetid.getncattr('calendar')

                    data = num2date(data,units=units,calendar=calendar)

                    return data

        except BaseException as e :
            print(e)
            return None

        return data


def writenc(filename, sdsname, srcdata, dimension=None, overwrite=1,
            complevel=9, dictsdsinfo=None, fill_value=None,
            standard_name=None, long_name=None, description=None, units=None,
            valid_range=None,
            scale_factor=None, add_offset=None, **kwargs):
    '''
    写NC文件

    Parameters
    ----------
    filename : str
    sdsname : str
    srcdata : str
    dimension : tuple
        数据集关联的维度名
    overwrite : int
    complevel : int
        压缩等级，0-9，值越大，压力等级越大
    dictsdsinfo ：dict
        数据集属性
    fill_value  : int or float
        填充值 或 无效值
    standard_name
    long_name
    description
    units
    valid_range
    scale_factor
    add_offset
    kwargs

    Returns
    -------

    '''

    data = np.array(srcdata)

    if (dictsdsinfo is None) or not isinstance(dictsdsinfo, dict) :
        dictsdsinfo={}

    dictsdsinfo = checkinfo(dictsdsinfo, 'standard_name', standard_name)
    dictsdsinfo = checkinfo(dictsdsinfo, 'standard_name', sdsname)
    dictsdsinfo = checkinfo(dictsdsinfo, 'long_name', long_name)
    dictsdsinfo = checkinfo(dictsdsinfo, 'long_name', sdsname)
    dictsdsinfo = checkinfo(dictsdsinfo, 'valid_range', valid_range)
    dictsdsinfo = checkinfo(dictsdsinfo, 'units', units)
    dictsdsinfo = checkinfo(dictsdsinfo, 'scale_factor', scale_factor)
    dictsdsinfo = checkinfo(dictsdsinfo, 'add_offset', add_offset)
    dictsdsinfo = checkinfo(dictsdsinfo, 'description', description)

    for key in kwargs :
        if not key in dictsdsinfo :
            dictsdsinfo[key] = kwargs[key]

    if overwrite :
        fp = netCDF4.Dataset(filename, 'w')
    else:
        fp = netCDF4.Dataset(filename, 'r+')

    dsetid = None
    # 如果没有对应的关联维度，则创建该维度
    if dimension is None :
        fp.createDimension(sdsname, data.shape[0])
        dsetid = fp.createVariable(sdsname, data.dtype, dimensions=sdsname,
                                  zlib=True, complevel=9)
        fp.variables[sdsname][:] = data
    else:
        # 判断数据集维度是否存在，如果不存在，则创建该维度，并赋值为0
        k = 0
        for item in dimension :
            if not item in fp.dimensions :
                fp.createDimension(item, data.shape[k])
                dsetid = fp.createVariable(item, np.int32, dimensions=item, zlib=True, complevel=9)
                fp.variables[item][:] = range(data.shape[k])
            k = k + 1
        dsetid = fp.createVariable(sdsname, data.dtype, dimensions=dimension, zlib=True,
                                   complevel=complevel, fill_value=fill_value)
        fp.variables[sdsname][:] = data

    if (not dictsdsinfo is None) and (not dsetid is None) :
        for key in dictsdsinfo :
            dsetid.setncattr(key, dictsdsinfo[key])

    fp.close()
    print('create %s %s success...' %(filename, sdsname))


def writenc_fileinfo(filename, dictfileinfo, overwrite=1):

    try:
        if overwrite :
            fp = netCDF4.Dataset(filename, 'w')
        else:
            if os.path.isfile(filename) :
                fp = netCDF4.Dataset(filename, 'r+')
            else:
                print('%s is not exist, will create it...' %(filename))
                fp = netCDF4.Dataset(filename, 'w')

        if not dictfileinfo is None :
            for key in dictfileinfo:
                fp.setncattr(key, dictfileinfo[key])
        fp.close()
        return True

    except BaseException as e :
        print('write %s file information error!!!' %(filename))
        return False


def writenc_sdsinfo(filename, sdsname, dictsdsinfo, overwrite=1):
    try:
        if overwrite:
            fp = netCDF4.Dataset(filename, 'w')
        else:
            if os.path.isfile(filename):
                fp = netCDF4.Dataset(filename, 'r+')
            else:
                print('%s is not exist, will create it...' % (filename))
                fp = netCDF4.Dataset(filename, 'w')
        if not sdsname in fp.variables:
            print('%s is not a dataset in %s' %(sdsname, filename))
            fp.close()
            return False
        else:
            dsetid = fp.variables[sdsname]
            for key in dictsdsinfo:
                dsetid.setncattr(key, dictsdsinfo[key])
        fp.close()
        return True
    except BaseException as e:
        print('write %s dataset %s information error!!!' % (filename, sdsname))
        return False

def writencfortimes(filename, sdsname, srcdata, overwrite=1,
                    units = 'hours since 1900-01-01 00:00:00.0',
                    calendar = "gregorian",
                    complevel=9, dictsdsinfo=None):
    '''
    写时间维度数据集

    Parameters
    ----------
    filename : str
    sdsname: str
        datetime array
    srcdata : datetime array

    overwrite: int
    units : str
        时间单位，从什么时间开始计数
    calendar : str
    complevel : int
    dictsdsinfo : dict
        数据集属性

    Returns
    -------

    '''

    data = np.array(srcdata)

    if overwrite :
        fp = netCDF4.Dataset(filename, 'w')
    else:
        fp = netCDF4.Dataset(filename, 'r+')

    # 如果没有对应的关联维度，则创建该维度

    fp.createDimension(sdsname, data.shape[0])
    times = fp.createVariable(sdsname, 'f8', dimensions=sdsname,
                              zlib=True, complevel=complevel)
    times.units = units
    times.calendar = calendar

    fp.variables[sdsname][:] = date2num(data,units=times.units,calendar=times.calendar)


    if (not dictsdsinfo is None) and (not times is None) :
        for key in dictsdsinfo :
            times.setncattr(key, dictsdsinfo[key])

    fp.close()
    print('create %s %s success...' %(filename, sdsname))


def checkinfo(dictinfo, key, value):

    if not key in dictinfo :
        if value is not None :
            dictinfo[key] = value

    return dictinfo




# -*- coding:utf-8 -*-
'''
@Project  : fypy

@File     : nc2tif.py

@Modify Time : 2022/8/11 15:34

@Author : Lee

@Version : 1.0

@Description :
针对netCDF4数据，将数据集进行投影转换成GeoTiff文件

'''
import os
import sys
import numpy as np

import netCDF4
from fypy.tools import writetiff


def nc2tif(outname, filename, sdsname, fillvalue=-999):
    '''
    将NC文件中的数据集根据lat\lon维度进行投影输出geotiff文件

    Parameters
    ----------
    outname: str
        输出tif文件名
    filename: str
        输入的nc文件名
    sdsname: str
        需要转换的数据集名
    fillvalue: float or int
        数据填充值
    Returns
    -------
        bool
        成功返回TRUE，失败返回FALSE
    '''

    try:
        fp = netCDF4.Dataset(filename, 'r')

        dsetid = fp[sdsname]
        dims = dsetid.dimensions

        grid = False
        xdim = None
        ydim = None
        for item in dims :
            if 'lon' in item.lower() :
                grid = True
                xdim = item

            if 'lat' in item.lower() :
                grid = True
                ydim = item

        sdsinfo = dsetid.ncattrs()
        if xdim is not None and ydim is not None :
            x = fp[xdim][:]
            y = fp[ydim][:]
            xres = x[1] - x[0]
            yres = y[1] - y[0]

            trans = [x[0], xres, 0, y[0], 0, yres]
            data = fp[sdsname][:]
            if '_FillValue' in sdsinfo :
                fillvalue = sdsinfo['_FillValue']
            writetiff(outname, data, im_geotrans=trans, fillvalue=fillvalue)
        fp.close()

        return True
    except BaseException as e :
        print(e)
        return False

    # 后续更新根据coordinate属性来完成投影设置

    for item in sdsinfo :
        if 'coordinate' in item.lower() :
            grid = True






# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : __init__.py

@Modify Time :  2023/4/27 14:46   

@Author      : Lee    

@Version     : 1.0   

@Description :

'''

from .AtmCorr_GF import AtmCorr_GF
from .AtmCorr_Landsat import AtmCorr_Landsat
from .AtmCorr_FY3D_MERSI import AtmCorr_FY3D_MERSI
from .AtmCorr_Sentinel import AtmCorr_Sentinel




# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : fy4searchtable.py

@Modify Time :  2022/11/10 14:45

@Author      : Lee

@Version     : 1.0

@Description :

'''
import os

import sys
import numpy as np
import datetime


TWOPI = 6.28318530717958648
DPAI = 6.28318530717958648

deg2rad = np.pi / 180.0
rad2deg = 180.0 / np.pi
FILLVALUE = -999


# 北边界纬度/º	80.56672132
# 南边界纬度/º	-80.56672132
# 东边界经度/º	-174.71662309
# 西边界经度/º	24.11662309


# 分辨率(km)   行数    列数
#   16        687    687
#   8        1374   1374
#   4        2748   2748
#   2        5496   5496
#   1       10992  10992
# 0.5       21984  21984
# 0.25      43968  43968

class fy4searchtable(object) :

    def __init__(self, resolution=0.04):

        self.ea = 6378.137          # 地球长半轴
        self.eb = 6356.7523         # 地球短半轴
        self.H = 42164.0        # 地心到卫星质心的距离

        if resolution == 0.0025 :  # 250米
            self.coff = 21983.5
            self.cfac = 163730199
            self.loff = 21983.5
            self.lfac = 163730199
            self.rowmax = 43968
            self.colmax = 43968
        elif resolution == 0.005 :
            self.coff = 10991.5
            self.cfac = 81865099
            self.loff = 10991.5
            self.lfac = 81865099
            self.rowmax = 21984
            self.colmax = 21984
        elif resolution == 0.01 :
            self.coff = 5495.5
            self.cfac = 40932549
            self.loff = 5495.5
            self.lfac = 40932549
            self.rowmax = 10992
            self.colmax = 10992
        elif resolution == 0.02 :
            self.coff = 2747.5
            self.cfac = 20466274
            self.loff = 2747.5
            self.lfac = 20466274
            self.rowmax = 5496
            self.colmax = 5496
        elif resolution == 0.04 :
            self.coff = 1373.5      # 1375.5
            self.cfac = 10233137    # 10233137
            self.loff = 1373.5      # 1375.5
            self.lfac = 10233137    # 10233137
            self.rowmax = 2748
            self.colmax = 2748
        else:
            raise Exception("resolution error! please input [0.005, 0.01, 0.02, 0.04]")


        self.FLAT = 0.00335281317789691
        self.AE = 6378.137
        self.E2 = 0.0066943800699785


    def IJ2LatLon(self, row, col, subpoint=104.7):
        '''
        行列号转经纬度

        Parameters
        ----------
        row:  numpy.array
            行号
        col:  numpy.array
            列号
        subpoint : float
            星下点（degree）

        Returns
        -------
            numpy.array
            返回纬度、经度
        '''

        row = np.array(row)
        col = np.array(col)

        fillflag = (row < 0) | (row >= self.rowmax) | (col < 0) | (col >= self.colmax)

        x = deg2rad * (col - self.coff) / (2**-16 * self.cfac)
        y = deg2rad * (row - self.loff) / (2**-16 * self.lfac)

        sd = (self.H * np.cos(x) * np.cos(y)) * (self.H * np.cos(x) * np.cos(y)) \
             - (np.cos(y) * np.cos(y) + (self.ea * self.ea) / (self.eb * self.eb) \
                * np.sin(y) * np.sin(y)) * ((self.H * self.H) - self.ea * self.ea)

        flag = sd < 0

        sd[sd>=0] = np.sqrt(sd[sd>=0])

        sn = (self.H * np.cos(x) * np.cos(y) - sd) / \
             (np.cos(y) * np.cos(y) + (self.ea * self.ea) / (self.eb * self.eb) * np.sin(y) * np.sin(y))

        S1 = self.H - (sn * np.cos(x) * np.cos(y))
        S2 = sn * np.sin(x) * np.cos(y)
        S3 = -sn * np.sin(y)
        Sxy = np.sqrt(S1 * S1 + S2 * S2)

        lon = rad2deg * np.arctan2(S2 , S1) + subpoint
        lat = rad2deg * np.arctan((self.ea * self.ea) / (self.eb * self.eb) * S3 / Sxy)

        lon[lon > 180] -= 360.0
        lon[lon < -180] += 360.0

        lon[flag] = FILLVALUE
        lat[flag] = FILLVALUE

        lon[fillflag] = FILLVALUE
        lat[fillflag] = FILLVALUE

        lat = np.array(lat, dtype=np.float32)
        lon = np.array(lon, dtype=np.float32)

        return  lat, lon


    def LatLon2IJ(self, lat, lon, subpoint=104.7):
        '''
        经纬度转行列号

        Parameters
        ----------
        lat : numpy.array
            纬度， -90.0 ~ 90.0
        lon : numpy.array
            经度， -180.0 ~ 180.0
        subpoint: float
            星下点（degree）
        Returns
        -------
             numpy.array
             返回行号、列号
        '''

        lat1 = lat.copy()
        lon1 = lon.copy()

        lon1[lon1 > 180] -= 360.0

        lon1 = lon * deg2rad
        lat1 = lat * deg2rad

        phi_e = np.arctan(self.eb ** 2 * np.tan(lat1) / self.ea ** 2)
        re = self.eb / np.sqrt(1-(self.ea**2 - self.eb**2) * np.cos(phi_e)**2/self.ea**2)
        r1 = self.H - re * np.cos(phi_e) * np.cos(lon1 - subpoint* deg2rad)
        r2 = -re * np.cos(phi_e) * np.sin(lon1 - subpoint* deg2rad)
        r3 = re * np.sin(phi_e)
        rn = np.sqrt(r1**2 + r2**2 + r3**2)
        x = np.arctan2(-r2, r1)*180/np.pi
        # x = np.arctan(-r2/r1)*180/np.pi
        y = np.arcsin(-r3/rn)*180/np.pi
        col = self.coff + x * 2**(-16) * self.cfac
        line = self.loff + y * 2**(-16) * self.lfac

        # 对外太空进行判断
        tmp = r1 * (r1 - self.H) + r2 **2 + r3**2
        flag = tmp > 0

        line = np.array(line+0.5, dtype=np.int32)
        col = np.array(col+0.5, dtype=np.int32)

        line[(line<0) | (line >= self.rowmax)] = FILLVALUE
        col[(col<0) | (col>=self.colmax)]  = FILLVALUE

        return line, col


    def CalSatZenith(self, lat, lon, subpoint = 104.7) :
        lat1 = np.array(lat, dtype=np.float32)
        lon1 = np.array(lon, dtype=np.float32)
        fillflag = (lat1 < -90) | (lat1 > 90) | (lon1 < -180) | (lon1 > 180)

        lat1 = lat1 * deg2rad
        lon1 = lon1 * deg2rad
        point = subpoint * deg2rad

        c_lat = np.arctan(0.993243 * np.tan(lat1))
        rl = 6356.7523 / np.sqrt(1 - 0.00675701 * np.cos(c_lat) * np.cos(c_lat))
        r1 = self.H - rl * np.cos(c_lat) * np.cos(lon1 - point)
        r2 = -rl * np.cos(c_lat) * np.sin(lon1 - point)

        r3 = rl * np.sin(c_lat)
        rn = np.sqrt(r1 * r1 + r2 * r2 + r3 * r3)

        r4 = np.arccos((rn * rn + self.H * self.H - rl * rl) / (2 * rn * self.H))

        r5 = np.arcsin(np.sin(r4) * self.H / 6356.7523) * rad2deg

        r5[fillflag] = FILLVALUE

        return r5


    def CalSatAzimuth(self, lat, lon, subpoint = 104.7):
        '''
        计算卫星方位角
        :param lat1: array_like,
                    unit: degree
        :param lon1: array_like,
                    unit: degree
        :param subpoint: float,
                    unit: degree
        :return: satellite azimuth , array_like:
                    unit:degree
        '''
        lat1 = np.array(lat, dtype=np.float32)
        lon1 = np.array(lon, dtype=np.float32)
        fillflag = (lat1 < -90) | (lat1 > 90) | (lon1 < -180) | (lon1 > 180)

        lat1 = lat1 * deg2rad
        lon1 = lon1 * deg2rad
        point = subpoint * deg2rad

        heigh = self.H * 1.0E3      # km -> m

        RX = self._Zocgef(lat1, lon1, heigh)

        RSat = self._Zocgef(0.0, point, heigh)

        Azi = self._ICGSSAT(lat1, lon1, RSat, RX)

        sata = Azi * rad2deg

        flag = sata >= 180
        sata[flag] -= 180.0
        sata[~flag] += 180.0

        sata[fillflag] = FILLVALUE

        return sata


    def CalSunZenith(self, lat, lon, nowtime):
        '''
        计算太阳天顶角
        :param lat: array_like,
                    unit: degree
        :param lon: array_like,
                    unit: degree
        :param nowtime: datetime
        :return: sun zenith , array_like:
                    unit:degree
        '''
        lat1 = np.array(lat, dtype=np.float32)
        lon1 = np.array(lon, dtype=np.float32)
        fillflag = (lat1 < -90) | (lat1 > 90) | (lon1 < -180) | (lon1 > 180)

        lat1 = lat1 * deg2rad
        lon1 = lon1 * deg2rad

        sttime = (nowtime - datetime.datetime.strptime('%s-01-01' % (nowtime.strftime('%Y')), '%Y-%m-%d')).total_seconds()

        PHIG,  DPHIDT, CHISUN,  EPSLON = self._EARTH(sttime)

        RSOL = 149600881.0
        COSX = np.cos(CHISUN)
        SINX = np.sin(CHISUN)
        COSE = np.cos(EPSLON)
        SINE = np.sin(EPSLON)

        RX = [RSOL*COSX,  RSOL*(SINX*COSE),  RSOL*(SINX*SINE)]

        TH = np.fmod(lon1 + PHIG, TWOPI)
        cTH = np.cos(TH)
        sTH = np.sin(TH)

        Azgs, Elgs = self._ICGS(RX, sTH, cTH, lat1)

        # SunAzi = Azgs * rad2deg
        SunZen = ((np.pi / 2.0) - Elgs) * rad2deg

        SunZen[fillflag] = FILLVALUE

        return SunZen


    def CalSunAzimuth(self, lat, lon, nowtime):

        lat1 = np.array(lat, dtype=np.float32)
        lon1 = np.array(lon, dtype=np.float32)
        fillflag = (lat1 < -90) | (lat1 > 90) | (lon1 < -180) | (lon1 > 180)

        lat1 = lat1 * deg2rad
        lon1 = lon1 * deg2rad

        sttime = (nowtime - datetime.datetime.strptime('%s-01-01' % (nowtime.strftime('%Y')), '%Y-%m-%d')).total_seconds()

        PHIG,  DPHIDT, CHISUN,  EPSLON = self._EARTH(sttime)

        RSOL = 149600881.0
        COSX = np.cos(CHISUN)
        SINX = np.sin(CHISUN)
        COSE = np.cos(EPSLON)
        SINE = np.sin(EPSLON)

        RX = [RSOL*COSX,  RSOL*(SINX*COSE),  RSOL*(SINX*SINE)]

        TH = np.fmod(lon1 + PHIG, TWOPI)
        cTH = np.cos(TH)
        sTH = np.sin(TH)

        Azgs, Elgs = self._ICGS(RX, sTH, cTH, lat1)

        SunAzi = Azgs * rad2deg
        # SunZen = ((np.pi / 2.0) - Elgs) * rad2deg

        SunAzi[fillflag] = FILLVALUE

        return SunAzi


    def CalSunGL(self, satz, sunz, rela):
        '''

        :param satz:
        :param sunz:
        :param rela:
        :return:
        '''
        fillflag = (satz < -360) | (satz > 360) | (sunz < -360) | (sunz > 360) | (rela < -360) | (rela > 360)

        data = np.cos(satz * deg2rad) * np.cos(sunz * deg2rad) \
               + np.sin(satz * deg2rad) * np.sin(sunz * deg2rad) * np.cos(rela * deg2rad)
        # cos_solzen * cos_satzen + sin_solzen * sin_satzen * cos_relaz

        data[data > 1.0] = 0.0
        SunGL = np.arccos(data) * rad2deg
        SunGL[fillflag] = -999.0

        return SunGL


    def CalRelativeAzimuth(self, sata, suna):
        fillflag = (sata > 360) | (sata < 0) | (suna > 360) | (suna < 0)

        RelativeAzi = np.fabs(sata - suna) + 180.0 # ???? + 180

        RelativeAzi[fillflag] = -999.0

        return RelativeAzi

    def _ICGSSAT(self, lat, lon, RSat, RX):

        HGT = 50.0 / 6378137.0
        Sla = np.sin(lat)
        Cla = np.cos(lat)
        Sth = np.sin(lon)
        Cth = np.cos(lon)

        Rhx = RSat[0] - RX[0]
        Rhy = RSat[1] - RX[1]
        Rhz = RSat[2] - RX[2]
        Rht = Cth * Rhx + Sth * Rhy
        Rxgs = Cth * Rhy - Sth * Rhx
        Rygs = Cla * Rhz - Sla * Rht
        Rzgs = Cla * Rht + Sla * Rhz

        Azgs = np.arctan2(Rxgs, Rygs)
        Azgs[Azgs < 0.0] = Azgs[Azgs < 0.0] + TWOPI
        Azgs[Azgs < np.pi] = Azgs[Azgs < np.pi] + np.pi
        Azgs[Azgs >= np.pi] = Azgs[Azgs >= np.pi] - np.pi

        Elgs = np.arctan(Rzgs / np.sqrt(Rxgs * Rxgs + Rygs * Rygs))

        return Elgs


    def _Zocgef(self, lat, lon, height):

        Xge = [] # np.zeros(3, dtype=np.float64)
        E = self.FLAT * (2.0 - self.FLAT)
        ENP = (self.AE * 1000) / np.sqrt(1.0 - E * np.sin(lat) * E * np.sin(lat))
        Xge.append( (ENP + height) * np.cos(lat) * np.cos(lon) / 1000.)

        Xge.append((ENP + height) * np.cos(lat) * np.sin(lon) / 1000.)

        Xge.append((ENP * (1.0 - E) + height) * np.sin(lat) / 1000.)

        return Xge


    def _ICGS(self, RX, Sth, Cth, lat):
        HGT = 50.0 / 6378137.0
        Sla = np.sin(lat)
        Cla = np.cos(lat)
        RE2 = self.AE / np.sqrt(1.0 - self.E2 * Sla * Sla)
        Gxic = (RE2 + HGT) * Cla * Cth

        Gyic = (RE2 + HGT) * Cla * Sth
        Gzic = (RE2 * (1.0 - self.E2) + HGT) * Sla
        Rhx = RX[0] - Gxic
        Rhy = RX[1] - Gyic
        Rhz = RX[2] - Gzic
        Rht = Cth * Rhx + Sth * Rhy
        Rxgs = Cth * Rhy - Sth * Rhx
        Rygs = Cla * Rhz - Sla * Rht
        Rzgs = Cla * Rht + Sla * Rhz

        Azgs = np.arctan2(Rxgs, Rygs)

        Azgs[Azgs < 0.0] = Azgs[Azgs < 0.0] + TWOPI
        Elgs = np.arctan(Rzgs / np.sqrt(Rxgs * Rxgs + Rygs * Rygs))

        return Azgs, Elgs

    def _EARTH(self, stime):
        PHIGRN = [ 1.739935890, 628.33195070, .000006754947810]
        CHIMS = [4.881633570, 628.3319487050, 1.4468088e-10]

        MEARTH = [6.256581470, 628.3019468590, -7.1676766e-11]
        OMEGA = [4.523636080, -33.7571619360, 9.9285594E-10]
        EPSLON = [.40936890950, - 2.27232172336e-4]
        KAPPA = .0335020
        DELPSI = [8.35465e-5,  .61712e-5]
        DELEPS = [4.46513e-5, .26771e-5]
        ZERO80 = 29219.50
        CENTCT = 36525.0

        DDATE = ZERO80 + stime / 86400.0
        IDATE2 = 2. * DDATE

        if IDATE2 % 2 == 0 :
            IDATE2 -= 1
        # IDATE2[(IDATE2 % 2) == 0] = IDATE2[(IDATE2 % 2) == 0] - 1
        DATE = IDATE2 * .50

        if DDATE - DATE > .50 :
            DATE += 1
        # DATE[DDATE - DATE > .50] = DATE[DDATE - DATE > .50] + 1.0
        EPTIME = (DATE - ZERO80) * 86400.0
        CENTNO = DATE / CENTCT

        ALPHA0 = self._ZSODSRadto2Pai(PHIGRN[0] + self._ZSODSRadto2Pai(CENTNO * (PHIGRN[1] + CENTNO * PHIGRN[2])))

        DADT = (TWOPI + (PHIGRN[1] + 2. * CENTNO * PHIGRN[2]) / CENTCT) / 86400.

        CHIMO = self._ZSODSRadto2Pai(CHIMS[0] + self._ZSODSRadto2Pai(CENTNO * (CHIMS[1] + CENTNO * CHIMS[2])))

        DXMDT = (CHIMS[1] + 2. * CENTNO * CHIMS[2]) / (86400.0 * CENTCT)

        MANOM0 = self._ZSODSRadto2Pai(MEARTH[0] + self._ZSODSRadto2Pai(CENTNO * (MEARTH[1] + CENTNO * MEARTH[2])))

        DMDT = (MEARTH[1] + 2. * CENTNO * MEARTH[2]) / (86400.0 * CENTCT)

        OMEGA0 = self._ZSODSRadto2Pai(OMEGA[0] + self._ZSODSRadto2Pai(CENTNO * (OMEGA[1] + CENTNO * OMEGA[2])))

        DWDT = (OMEGA[1] + 2. * CENTNO * OMEGA[2]) / (86400.0 * CENTCT)

        EPSLN0 = EPSLON[0] + CENTNO * EPSLON[1]

        DEDT = EPSLON[1] / (86400.0 * CENTCT)

        DELP0 = DELPSI[0] * np.sin(OMEGA0) + DELPSI[1] * np.sin(2.0 * CHIMO)

        DDPDT = DELPSI[0] * np.cos(OMEGA0) * DWDT + DELPSI[1] * np.cos(2. * CHIMO) *2.0 * DXMDT

        DELE0 = DELEPS[0] * np.cos(OMEGA0) + DELEPS[1] * np.cos(2.0 * CHIMO)

        DDEDT = -DELEPS[0] * np.sin(OMEGA0) * DWDT - DELEPS[1] * np.sin(2.0 * CHIMO) * 2.0 * DXMDT

        CHI0 = CHIMO + KAPPA * np.sin(MANOM0)

        DXDT = DXMDT + KAPPA * np.cos(MANOM0) * DMDT

        ANUT0 = DELP0 * np.cos(EPSLN0 + DELE0)

        DANDT = DDPDT * ANUT0 / DELP0 - DELP0 * np.sin(EPSLN0 + DELE0) * (DEDT + DDEDT)

        Dtime = stime - EPTIME

        ALPHA = self._ZSODSRadto2Pai(ALPHA0 + ANUT0 + (DADT + DANDT) * Dtime)

        ALPDOT = DADT + DANDT

        CHISUN = self._ZSODSRadto2Pai(CHI0 + DXDT * Dtime)

        OBLIQE = EPSLN0 + DELE0 + (DEDT + DDEDT) * Dtime

        return ALPHA, ALPDOT, CHISUN, OBLIQE


    def _ZSODSRadto2Pai(self, dRad) :

        dresult = dRad - DPAI*((int)(dRad/DPAI))

        if dRad < 0.0 :
            dresult += DPAI

        return dresult


def ijtolatlon(line, col, subpoint=104.7, resolution=0.04):

    myfy4a = fy4searchtable(resolution)

    lat, lon = myfy4a.IJ2LatLon(line, col, subpoint=subpoint)

    return lat, lon

def latlon2ij(lat, lon, subpoint=104.7, resolution=0.04):

    myfy4a = fy4searchtable(resolution)

    line, col = myfy4a.LatLon2IJ(lat, lon, subpoint=subpoint)

    return line, col









# -*- coding:utf-8 -*-
'''
@Project  : fypy

@File     : hdfpro.py

@Modify Time : 2022/8/11 15:34

@Author : Lee

@Version : 1.0

@Description :
对hdf5文件进行读写操作

'''
import os
import h5py

def readhdf(filename, sdsname, dictsdsinfo=None) :
    r'''
    读取hdf数据集和属性
    :param filename:
    :param sdsname:
    :param dictsdsinfo:
    :return:
    '''
    if not os.path.isfile(filename) :
        print('%s is not exist, will return None' %(filename))
        return None
    else:
        try:
            fp = h5py.File(filename, 'r')
            dsetid = fp[sdsname]
    
            data = dsetid[:]
            if not dictsdsinfo is None :
                for key in dsetid.attrs :
                    # print(key)
                    dictsdsinfo.update({key : dsetid.attrs[key]})
                fp.close()
                return data, dictsdsinfo
    
            fp.close()
        except BaseException as e :
            print(e)
            return None

        return data

def readhdf_fileinfo(filename) :
    '''
    读取hdf数据集和属性
    :param filename:
    :param sdsname:
    :param dictsdsinfo:
    :param dictfileinfo:
    :return:
    '''

    dictfileinfo = {}
    if not os.path.isfile(filename) :
        print('%s is not exist, will return None' %(filename))
        return dictfileinfo
    else:
        fp = h5py.File(filename, 'r')
        for key in fp.attrs :
            # print(key)
            dictfileinfo.update({key : fp.attrs[key]})
        fp.close()
        return dictfileinfo

def writehdf(filename, sdsname, data, overwrite=True, dictsdsinfo=None, dictfileinfo=None,
             compression=9, info = False):
    '''
    创建hdf5文件
    :param filename:
    :param sdsname:
    :param data:
    :param overwrite:
    :param dictsdsinfo:
    :param dictfileinfo:
    :param compression:
    :return:
    '''

    try:

        if overwrite :
            fp = h5py.File(filename, 'w')
        else:
            fp = h5py.File(filename, 'r+')
        if not dictfileinfo is None :
            for key in dictfileinfo :
                fp.attrs[key] = dictfileinfo[key]

        dsetid = fp.create_dataset(sdsname, data=data, compression = compression)

        if not dictsdsinfo is None :
            for key in dictsdsinfo :
                dsetid.attrs[key] = dictsdsinfo[key]

        fp.close()
        if info:
            print('create %s %s success...' %(filename, sdsname))
    except BaseException as e :
        print(e)
        return False

    return True


def writehdf_fileinfo(filename, dictfileinfo, overwrite=True, info=False):
    '''
    创建文件属性
    :param filename:
    :param dictfileinfo:
    :param overwrite:
    :param info: 是否打印信息
    :return:
    '''

    try:

        if overwrite :
            fp = h5py.File(filename, 'w')
        else:
            fp = h5py.File(filename, 'r+')

        if isinstance(dictfileinfo, dict):
            for key in dictfileinfo :
                fp.attrs[key] = dictfileinfo[key]
                if info :
                    print('create %s attr [%s] success...' %(
                        filename, key
                    ))


    except BaseException as e :
        print(e)
        return False

    return True




# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : __init__.py

@Modify Time :  2022/11/10 14:35   

@Author      : Lee    

@Version     : 1.0   

@Description :

'''
from .drawH8Image import drawH8TrueColor
from .ahi8_l1_pro import ahi8_l1_pro, hsd2hdf
from .ahi8searchtable import ahi8searchtable, ijtolatlon, latlon2ij


# -*- coding:utf-8 -*-
'''
@Project     : fypy

@File        : version.py

@Modify Time :  2022/11/11 14:09   

@Author      : Lee    

@Version     : 1.0   

@Description :

'''

import json

version_json = '''
{
 "version": "1.0.0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)

# file generated by setuptools_scm
# don't change, don't track in version control
# version = '1.1.13'fff
# version_tuple = (1, 1, 13)